"""
Utilities
=========

Package-wide utilities.

"""
from numbers import Number
from typing import Any, Callable, Optional, Tuple, Union

import numpy as np
from scipy import cluster


def _check_none(fun: Callable[..., Any]) -> Callable[..., Union[Any, None]]:
    """Return None if any input argument is None; Wrapper function."""

    def checked_fun(*args, **kwargs):
        if any(x is None for x in [*args, *(kwargs.values())]):
            return None
        return fun(*args, **kwargs)

    return checked_fun


@_check_none
def res_to_chi2(res: np.ndarray) -> float:
    """Translate residuals to chi2 values, `chi2 = sum(res**2) + C`."""
    return float(np.dot(res, res))


@_check_none
def chi2_to_fval(chi2: float) -> float:
    """Translate chi2 to function value, `fval = 0.5*chi2 = 0.5*sum(res**2) + C`.

    Note that for the function value we thus employ a probabilistic
    interpretation, as the log-likelihood of a standard normal noise model.
    This is in line with e.g. AMICI's and SciPy's objective definition.
    """
    return 0.5 * chi2


@_check_none
def fval_to_chi2(fval: float) -> float:
    """Translate function value to chi2, `chi2 = 2 * fval`.

    Note that for the function value we thus employ a probabilistic
    interpretation, as the log-likelihood of a standard normal noise model.
    This is in line with e.g. AMICI's and SciPy's objective definition.
    """
    return 2.0 * fval


@_check_none
def res_to_fval(res: np.ndarray) -> float:
    """Translate residuals to function value, `fval = 0.5*sum(res**2) + C`."""
    return chi2_to_fval(res_to_chi2(res))


@_check_none
def sres_to_schi2(res: np.ndarray, sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to chi2 gradient."""
    return 2 * res.dot(sres)


@_check_none
def schi2_to_grad(schi2: np.ndarray) -> np.ndarray:
    """Translate chi2 gradient to function value gradient.

    See also :func:`chi2_to_fval`.
    """
    return 0.5 * schi2


@_check_none
def grad_to_schi2(grad: np.ndarray) -> np.ndarray:
    """Translate function value gradient to chi2 gradient."""
    return 2.0 * grad


@_check_none
def sres_to_grad(res: np.ndarray, sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to function value gradient.

    Assumes `fval = 0.5*sum(res**2)`.

    See also :func:`chi2_to_fval`.
    """
    return schi2_to_grad(sres_to_schi2(res, sres))


@_check_none
def sres_to_fim(sres: np.ndarray) -> np.ndarray:
    """Translate residual sensitivities to FIM.

    The FIM is based on the function values, not chi2, i.e. has a normalization
    of 0.5 as in :func:`res_to_fval`.
    """
    return sres.transpose().dot(sres)


def is_none_or_nan(x: Union[Number, None]) -> bool:
    """
    Check if x is None or NaN.

    Parameters
    ----------
    x:
        object to be checked

    Returns
    -------
    True if x is None or NaN, False otherwise.
    """
    return x is None or np.isnan(x)


def is_none_or_nan_array(x: Union[Number, np.ndarray, None]) -> bool:
    """
    Check if x is None or NaN array.

    Parameters
    ----------
    x:
        object to be checked

    Returns
    -------
    True if x is None or NaN array, False otherwise.
    """
    return x is None or np.isnan(x).all()


def allclose(
    x: Union[Number, np.ndarray], y: Union[Number, np.ndarray]
) -> bool:
    """
    Check if two arrays are close.

    Parameters
    ----------
    x: first array
    y: second array

    Returns
    -------
    True if all elements of x and y are close, False otherwise.
    """
    # Note: We use this wrapper around np.allclose in order to more easily
    #  adjust hyper parameters for the tolerance.
    return np.allclose(x, y)


def isclose(
    x: Union[Number, np.ndarray],
    y: Union[Number, np.ndarray],
) -> Union[bool, np.ndarray]:
    """
    Check if two values or arrays are close, element-wise.

    Parameters
    ----------
    x: first array
    y: second array

    Returns
    -------
    Element-wise boolean comparison of x and y.
    """
    # Note: We use this wrapper around np.isclose in order to more easily
    #  adjust hyper parameters for the tolerance.
    return np.isclose(x, y)


def get_condition_label(condition_id: str) -> str:
    """Convert a condition ID to a label.

    Labels for conditions are used at different locations (e.g. ensemble
    prediction code, and visualization code). This method ensures that the same
    condition is labeled identically everywhere.

    Parameters
    ----------
    condition_id:
        The condition ID that will be used to generate a label.

    Returns
    -------
    The condition label.
    """
    return f'condition_{condition_id}'


def assign_clusters(vals):
    """
    Find clustering.

    Parameters
    ----------
    vals: numeric list or array
        List to be clustered.

    Returns
    -------
    clust: numeric list
        Indicating the corresponding cluster of each element from
        'vals'.
    clustsize: numeric list
        Size of clusters, length equals number of clusters.
    """
    # sanity checks
    if vals is None or len(vals) == 0:
        return [], []
    elif len(vals) == 1:
        return np.array([0]), np.array([1.0])

    # linkage requires (n, 1) data array
    vals = np.reshape(vals, (-1, 1))

    # however: clusters are sorted by size, not by value... Resort.
    # Create preallocated object first
    cluster_indices = np.zeros(vals.size, dtype=int)

    # get clustering based on distance
    clust = cluster.hierarchy.fcluster(
        cluster.hierarchy.linkage(vals), t=0.1, criterion='distance'
    )

    # get unique clusters
    _, ind_clust = np.unique(clust, return_index=True)
    unique_clust = clust[np.sort(ind_clust)]
    cluster_size = np.zeros(unique_clust.size, dtype=int)

    # loop over clusters: resort and count number of entries
    for index, i_clust in enumerate(unique_clust):
        cluster_indices[np.where(clust == i_clust)] = index
        cluster_size[index] = sum(clust == i_clust)

    return cluster_indices, cluster_size


def delete_nan_inf(
    fvals: np.ndarray, x: Optional[np.ndarray] = None, xdim: Optional[int] = 1
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Delete nan and inf values in fvals.

    If parameters 'x' are passed, also the corresponding entries are deleted.

    Parameters
    ----------
    x:
        array of parameters
    fvals:
        array of fval
    xdim:
        dimension of x, in case x dimension cannot be inferred

    Returns
    -------
    x:
        array of parameters without nan or inf
    fvals:
        array of fval without nan or inf
    """
    fvals = np.asarray(fvals)
    if x is not None:
        # if we start out with a list of x, the x corresponding
        # to finite fvals may be None, so we cannot stack the x before taking
        # subindexing
        # If none of the fvals are finite, np.vstack will fail and np.take
        # will not yield the correct dimension, so we try to construct an
        # empty np.ndarray with the correct dimension (other functions rely
        # on x.shape[1] to be of correct dimension)
        if np.isfinite(fvals).any():
            x = np.vstack(np.take(x, np.where(np.isfinite(fvals))[0], axis=0))
        else:
            x = np.empty(
                (
                    0,
                    x.shape[1]
                    if x.ndim == 2
                    else x[0].shape[0]
                    if x[0] is not None
                    else xdim,
                )
            )
    return x, fvals[np.isfinite(fvals)]



# noqa: D400,D205
"""
pyPESTO
=======

Parameter Estimation TOolbox for python.
"""

# isort: off

# make version available
from .version import __version__

# import basic objects into global namespace
from .history import (
    CountHistory,
    CountHistoryBase,
    CsvHistory,
    Hdf5History,
    NoHistory,
    HistoryBase,
    HistoryOptions,
    MemoryHistory,
    OptimizerHistory,
)
from .objective import (
    AmiciObjective,
    NegLogPriors,
    Objective,
    ObjectiveBase,
    FD,
    FDDelta,
)
from .problem import Problem
from .result import (
    McmcPtResult,
    OptimizeResult,
    OptimizerResult,
    ProfileResult,
    ProfilerResult,
    PredictionResult,
    PredictionConditionResult,
    Result,
    SampleResult,
)

# import simple modules as submodules
from . import (
    engine,
    logging,
    startpoint,
    store,
    visualize,
    C,
)

# isort: on

logging.log()



"""
Result
======

The pypesto.Result object contains all results generated by
the pypesto components. It contains sub-results for
optimization, profiling, sampling.
"""

from .optimize import OptimizeResult, OptimizerResult
from .predict import PredictionConditionResult, PredictionResult
from .profile import ProfileResult, ProfilerResult
from .result import Result
from .sample import McmcPtResult, SampleResult



"""
Model Selection
===============

Perform model selection with a PEtab Select problem.
"""

from . import postprocessors
from .misc import model_to_pypesto_problem
from .problem import Problem

try:
    import petab_select

    del petab_select
except ImportError:
    import warnings

    warnings.warn(
        "pyPESTO's model selection methods require an installation of PEtab "
        "Select (https://github.com/PEtab-dev/petab_select). Install via "
        "`pip3 install petab-select` or `pip3 install pypesto[select]`."
    )



import copy
from typing import Iterable, List, Optional, SupportsFloat, SupportsInt, Union

import numpy as np
import pandas as pd

from ..objective import ObjectiveBase
from ..objective.priors import NegLogParameterPriors

SupportsFloatIterableOrValue = Union[Iterable[SupportsFloat], SupportsFloat]
SupportsIntIterableOrValue = Union[Iterable[SupportsInt], SupportsInt]


class Problem:
    """
    The problem formulation.

    A problem specifies the objective function, boundaries and constraints,
    parameter guesses as well as the parameters which are to be optimized.

    Parameters
    ----------
    objective:
        The objective function for minimization. Note that a shallow copy
        is created.
    lb, ub:
        The lower and upper bounds for optimization. For unbounded directions
        set to +-inf.
    lb_init, ub_init:
        The lower and upper bounds for initialization, typically for defining
        search start points.
        If not set, set to lb, ub.
    dim_full:
        The full dimension of the problem, including fixed parameters.
    x_fixed_indices:
        Vector containing the indices (zero-based) of parameter components
        that are not to be optimized.
    x_fixed_vals:
        Vector of the same length as x_fixed_indices, containing the values
        of the fixed parameters.
    x_guesses:
        Guesses for the parameter values, shape (g, dim), where g denotes the
        number of guesses. These are used as start points in the optimization.
    x_names:
        Parameter names that can be optionally used e.g. in visualizations.
        If objective.get_x_names() is not None, those values are used,
        else the values specified here are used if not None, otherwise
        the variable names are set to ['x0', ... 'x{dim_full}']. The list
        must always be of length dim_full.
    x_scales:
        Parameter scales can be optionally given and are used e.g. in
        visualisation and prior generation. Currently the scales 'lin',
        'log`and 'log10' are supported.
    x_priors_defs:
        Definitions of priors for parameters. Types of priors, and their
        required and optional parameters, are described in the `Prior` class.
    copy_objective:
        Whethter to generate a deep copy of the objective function before
        potential modification the problem class performs on it.

    Notes
    -----
    On the fixing of parameter values:

    The number of parameters dim_full the objective takes as input must
    be known, so it must be either lb a vector of that size, or dim_full
    specified as a parameter.

    All vectors are mapped to the reduced space of dimension dim in __init__,
    regardless of whether they were in dimension dim or dim_full before. If
    the full representation is needed, the methods get_full_vector() and
    get_full_matrix() can be used.
    """

    def __init__(
        self,
        objective: ObjectiveBase,
        lb: Union[np.ndarray, List[float]],
        ub: Union[np.ndarray, List[float]],
        dim_full: Optional[int] = None,
        x_fixed_indices: Optional[SupportsIntIterableOrValue] = None,
        x_fixed_vals: Optional[SupportsFloatIterableOrValue] = None,
        x_guesses: Optional[Iterable[float]] = None,
        x_names: Optional[Iterable[str]] = None,
        x_scales: Optional[Iterable[str]] = None,
        x_priors_defs: Optional[NegLogParameterPriors] = None,
        lb_init: Union[np.ndarray, List[float], None] = None,
        ub_init: Union[np.ndarray, List[float], None] = None,
        copy_objective: bool = True,
    ):
        if copy_objective:
            objective = copy.deepcopy(objective)
        self.objective = objective

        self.lb_full: np.ndarray = np.array(lb).flatten()
        self.ub_full: np.ndarray = np.array(ub).flatten()
        if lb_init is None:
            lb_init = lb
        self.lb_init_full: np.ndarray = np.array(lb_init).flatten()
        if ub_init is None:
            ub_init = ub
        self.ub_init_full: np.ndarray = np.array(ub_init).flatten()

        self.dim_full: int = (
            dim_full if dim_full is not None else self.lb_full.size
        )

        if x_fixed_indices is None:
            x_fixed_indices = []
        x_fixed_indices = _make_iterable_if_value(x_fixed_indices, 'int')
        self.x_fixed_indices: List[int] = [
            _type_conversion_with_check(idx, ix, 'fixed indices', 'int')
            for idx, ix in enumerate(x_fixed_indices)
        ]

        # We want the fixed values to be a list, since we might need to add
        # or remove values during profile computation
        if x_fixed_vals is None:
            x_fixed_vals = []
        x_fixed_vals = _make_iterable_if_value(x_fixed_vals, 'float')
        self.x_fixed_vals: List[float] = [
            _type_conversion_with_check(idx, x, 'fixed values', 'float')
            for idx, x in enumerate(x_fixed_vals)
        ]

        if x_guesses is None:
            x_guesses = np.zeros((0, self.dim_full))
        self.x_guesses_full: np.ndarray = np.array(x_guesses)

        if x_names is None and objective.x_names is not None:
            x_names = objective.x_names
        elif x_names is None:
            x_names = [f'x{j}' for j in range(0, self.dim_full)]
        if len(set(x_names)) != len(x_names):
            raise ValueError("Parameter names x_names must be unique")
        self.x_names: List[str] = list(x_names)

        if x_scales is None:
            x_scales = ['lin'] * self.dim_full
        self.x_scales = x_scales

        self.x_priors = x_priors_defs

        self.normalize()

    @property
    def lb(self) -> np.ndarray:
        """Return lower bounds of free parameters."""
        return self.lb_full[self.x_free_indices]

    @property
    def ub(self) -> np.ndarray:
        """Return upper bounds of free parameters."""
        return self.ub_full[self.x_free_indices]

    @property
    def lb_init(self) -> np.ndarray:
        """Return initial lower bounds of free parameters."""
        return self.lb_init_full[self.x_free_indices]

    @property
    def ub_init(self) -> np.ndarray:
        """Return initial upper bounds of free parameters."""
        return self.ub_init_full[self.x_free_indices]

    @property
    def x_guesses(self) -> np.ndarray:
        """Return guesses of the free parameter values."""
        return self.x_guesses_full[:, self.x_free_indices]

    @property
    def dim(self) -> int:
        """Return dimension only considering non fixed parameters."""
        return self.dim_full - len(self.x_fixed_indices)

    @property
    def x_free_indices(self) -> List[int]:
        """Return non fixed parameters."""
        return sorted(set(range(0, self.dim_full)) - set(self.x_fixed_indices))

    def normalize(self) -> None:
        """
        Process vectors.

        Reduce all vectors to dimension dim and have the objective accept
        vectors of dimension dim.
        """
        for attr in ['lb_full', 'lb_init_full', 'ub_full', 'ub_init_full']:
            value = self.__getattribute__(attr)
            if value.size == 1:
                self.__setattr__(attr, value * np.ones(self.dim_full))
            elif value.size == self.dim:
                # in this case the bounds only holds the values of the
                # reduced bounds.
                self.__setattr__(
                    attr, self.get_full_vector(value, self.x_fixed_vals)
                )

            if self.__getattribute__(attr).size != self.dim_full:
                raise AssertionError(f"{attr} dimension invalid.")

        if self.x_guesses_full.shape[1] != self.dim_full:
            x_guesses_full = np.empty(
                (self.x_guesses_full.shape[0], self.dim_full)
            )
            x_guesses_full[:] = np.nan
            x_guesses_full[:, self.x_free_indices] = self.x_guesses_full
            self.x_guesses_full = x_guesses_full

        # make objective aware of fixed parameters
        self.objective.update_from_problem(
            dim_full=self.dim_full,
            x_free_indices=self.x_free_indices,
            x_fixed_indices=self.x_fixed_indices,
            x_fixed_vals=self.x_fixed_vals,
        )

        # sanity checks
        if len(self.x_scales) != self.dim_full:
            raise AssertionError("x_scales dimension invalid.")
        if len(self.x_names) != self.dim_full:
            raise AssertionError("x_names must be of length dim_full.")
        if len(self.x_fixed_indices) != len(self.x_fixed_vals):
            raise AssertionError(
                "x_fixed_indices and x_fixed_vals must have the same length."
            )
        if np.isnan(self.lb).any():
            raise ValueError('lb must not contain nan values')
        if np.isnan(self.ub).any():
            raise ValueError('ub must not contain nan values')
        if np.any(self.lb >= self.ub):
            raise ValueError('lb<ub not fulfilled.')

    def set_x_guesses(self, x_guesses: Iterable[float]):
        """
        Set the x_guesses of a problem.

        Parameters
        ----------
        x_guesses:
        """
        x_guesses_full = np.array(x_guesses)
        if x_guesses_full.shape[1] != self.dim_full:
            raise ValueError(
                'The dimension of individual x_guesses must be ' 'dim_full.'
            )
        self.x_guesses_full = x_guesses_full

    def fix_parameters(
        self,
        parameter_indices: SupportsIntIterableOrValue,
        parameter_vals: SupportsFloatIterableOrValue,
    ) -> None:
        """Fix specified parameters to specified values."""
        parameter_indices = _make_iterable_if_value(parameter_indices, 'int')
        parameter_vals = _make_iterable_if_value(parameter_vals, 'float')

        # first clean to-be-fixed indices to avoid redundancies
        for iter_index, (x_index, x_value) in enumerate(
            zip(parameter_indices, parameter_vals)
        ):
            # check if parameter was already fixed, otherwise add it to the
            # fixed parameters
            index = _type_conversion_with_check(
                iter_index, x_index, 'indices', 'int'
            )
            val = _type_conversion_with_check(
                iter_index, x_value, 'values', 'float'
            )
            if index in self.x_fixed_indices:
                self.x_fixed_vals[self.x_fixed_indices.index(index)] = val
            else:
                self.x_fixed_indices.append(index)
                self.x_fixed_vals.append(val)

        self.normalize()

    def unfix_parameters(
        self, parameter_indices: SupportsIntIterableOrValue
    ) -> None:
        """Free specified parameters."""
        # check and adapt input
        parameter_indices = _make_iterable_if_value(parameter_indices, 'int')

        # first clean to-be-freed indices
        for iter_index, x_index in enumerate(parameter_indices):
            index = _type_conversion_with_check(
                iter_index, x_index, 'indices', 'int'
            )
            if index in self.x_fixed_indices:
                fixed_x_index = self.x_fixed_indices.index(index)
                self.x_fixed_indices.pop(fixed_x_index)
                self.x_fixed_vals.pop(fixed_x_index)

        self.normalize()

    def get_full_vector(
        self, x: Union[np.ndarray, None], x_fixed_vals: Iterable[float] = None
    ) -> Union[np.ndarray, None]:
        """
        Map vector from dim to dim_full. Usually used for x, grad.

        Parameters
        ----------
        x: array_like, shape=(dim,)
            The vector in dimension dim.
        x_fixed_vals: array_like, ndim=1, optional
            The values to be used for the fixed indices. If None, then nans are
            inserted. Usually, None will be used for grad and
            problem.x_fixed_vals for x.
        """
        if x is None:
            return None

        # make sure it is an array
        x = np.array(x)

        if len(x) == self.dim_full:
            return x

        # Note: The funny indexing construct is to handle residual gradients,
        # where the last dimension is assumed to be the parameter one.
        x_full = np.zeros(x.shape[:-1] + (self.dim_full,))
        x_full[:] = np.nan
        x_full[..., self.x_free_indices] = x
        if x_fixed_vals is not None:
            x_full[..., self.x_fixed_indices] = x_fixed_vals
        return x_full

    def get_full_matrix(
        self, x: Union[np.ndarray, None]
    ) -> Union[np.ndarray, None]:
        """
        Map matrix from dim to dim_full. Usually used for hessian.

        Parameters
        ----------
        x: array_like, shape=(dim, dim)
            The matrix in dimension dim.
        """
        if x is None:
            return None

        # make sure it is an array
        x = np.array(x)

        if len(x) == self.dim_full:
            return x

        x_full = np.zeros((self.dim_full, self.dim_full))
        x_full[:, :] = np.nan
        x_full[np.ix_(self.x_free_indices, self.x_free_indices)] = x

        return x_full

    def get_reduced_vector(
        self,
        x_full: Union[np.ndarray, None],
        x_indices: Optional[List[int]] = None,
    ) -> Union[np.ndarray, None]:
        """
        Keep only those elements, which indices are specified in x_indices.

        If x_indices is not provided, delete fixed indices.

        Parameters
        ----------
        x_full: array_like, ndim=1
            The vector in dimension dim_full.
        x_indices:
            indices of x_full that should remain
        """
        if x_full is None:
            return None

        if x_indices is None:
            x_indices = self.x_free_indices

        if len(x_full) == len(x_indices):
            return x_full

        x = [x_full[idx] for idx in x_indices]
        return np.array(x)

    def get_reduced_matrix(
        self, x_full: Union[np.ndarray, None]
    ) -> Union[np.ndarray, None]:
        """
        Map matrix from dim_full to dim, i.e. delete fixed indices.

        Parameters
        ----------
        x_full: array_like, ndim=2
            The matrix in dimension dim_full.
        """
        if x_full is None:
            return None

        if len(x_full) == self.dim:
            return x_full

        x = x_full[np.ix_(self.x_free_indices, self.x_free_indices)]

        return x

    def full_index_to_free_index(self, full_index: int):
        """
        Calculate index in reduced vector from index in full vector.

        Parameters
        ----------
        full_index: The index in the full vector.

        Returns
        -------
        free_index: The index in the free vector.
        """
        fixed_indices = np.asarray(self.x_fixed_indices)
        if full_index in fixed_indices:
            raise ValueError(
                "Cannot compute index in free vector: Index is fixed."
            )
        return full_index - sum(fixed_indices < full_index)

    def print_parameter_summary(self) -> None:
        """
        Print a summary of parameters.

        Include what parameters are being optimized and parameter boundaries.
        """
        print(  # noqa: T201 (print)
            pd.DataFrame(
                index=self.x_names,
                data={
                    'free': [
                        idx in self.x_free_indices
                        for idx in range(self.dim_full)
                    ],
                    'lb_full': self.lb_full,
                    'ub_full': self.ub_full,
                },
            )
        )


_convtypes = {
    'float': {'attr': '__float__', 'conv': float},
    'int': {'attr': '__int__', 'conv': int},
}


def _type_conversion_with_check(
    index: int,
    value: Union[SupportsFloat, SupportsInt],
    valuename: str,
    convtype: str,
) -> Union[float, int]:
    """
    Convert values to the requested type.

    Raises and appropriate error if not possible.
    """
    if convtype not in _convtypes:
        raise ValueError(f'Unsupported type {convtype}')

    can_convert = hasattr(value, _convtypes[convtype]['attr'])
    # this may fail for weird custom ypes that can be converted to int but
    # not float, but we probably don't want those as indiced anyways
    lossless_conversion = not convtype == 'int' or (
        hasattr(value, _convtypes['float']['attr'])
        and (float(value) - int(value) == 0.0)
    )

    if not can_convert or not lossless_conversion:
        raise ValueError(
            f'All {valuename} must support lossless conversion to {convtype}. '
            f'Found type {type(value)} at index {index}, which cannot '
            f'be converted to {convtype}.'
        )

    return _convtypes[convtype]['conv'](value)


def _make_iterable_if_value(
    value: Union[SupportsFloatIterableOrValue, SupportsIntIterableOrValue],
    convtype: str,
) -> Union[Iterable[SupportsFloat], Iterable[SupportsInt]]:
    """Convert scalar values to iterables for scalar input, may update type."""
    if convtype not in _convtypes:
        raise ValueError(f'Unsupported type {convtype}')

    if not hasattr(value, '__iter__'):
        return [_type_conversion_with_check(0, value, 'values', convtype)]
    else:
        return value



from __future__ import annotations

import logging
from functools import partial
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
from scipy.stats import chi2

from ..C import (
    ENSEMBLE_TYPE,
    HISTORY,
    LOWER_BOUND,
    MEAN,
    MEDIAN,
    MODE_FUN,
    NVECTORS,
    NX,
    OUTPUT,
    OUTPUT_SENSI,
    PERCENTILE,
    PREDICTION_ARRAYS,
    PREDICTION_ID,
    PREDICTION_RESULTS,
    PREDICTION_SUMMARY,
    PREDICTIONS,
    PREDICTOR,
    STANDARD_DEVIATION,
    SUMMARY,
    TIMEPOINTS,
    UPPER_BOUND,
    VECTOR_TAGS,
    WEIGHTED_SIGMA,
    X_NAMES,
    X_VECTOR,
    EnsembleType,
    ModeType,
)
from ..engine import (
    Engine,
    MultiProcessEngine,
    MultiThreadEngine,
    SingleCoreEngine,
)

if TYPE_CHECKING:
    from ..objective import AmiciObjective

from ..result import PredictionConditionResult, PredictionResult, Result
from ..sample import geweke_test
from .task import EnsembleTask

logger = logging.getLogger(__name__)


class EnsemblePrediction:
    """
    Class of ensemble prediction.

    An ensemble prediction consists of an ensemble, i.e., a set of parameter
    vectors and their identifiers such as a sample, and a prediction function.
    It can be attached to a ensemble-type object
    """

    def __init__(
        self,
        predictor: Optional[Callable[[Sequence], PredictionResult]] = None,
        prediction_id: str = None,
        prediction_results: Sequence[PredictionResult] = None,
        lower_bound: Sequence[np.ndarray] = None,
        upper_bound: Sequence[np.ndarray] = None,
    ):
        """
        Initialize.

        Parameters
        ----------
        predictor:
            Prediction function, e.g., an AmiciPredictor, which takes a
            parameter vector as input and outputs a PredictionResult object
        prediction_id:
            Identifier for the predictions
        prediction_results:
            List of Prediction results
        lower_bound:
            Array of potential lower bounds for the predictions, should have
            the same shape as the output of the predictions, i.e., a list of
            numpy array (one list entry per condition), with the arrays having
            the shape of n_timepoints x n_outputs for each condition.
        upper_bound:
            array of potential upper bounds for the parameters
        """
        self.predictor = predictor
        if predictor is None:
            logger.info("This `EnsemblePrediction` has no predictor.")
        self.prediction_id = prediction_id
        self.prediction_results = prediction_results
        if prediction_results is None:
            self.prediction_results = []

        # handle bounds, Not yet Implemented
        if lower_bound is not None:
            raise NotImplementedError
        if upper_bound is not None:
            raise NotImplementedError
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        self.prediction_arrays = None
        self.prediction_summary = {
            MEAN: None,
            STANDARD_DEVIATION: None,
            MEDIAN: None,
            WEIGHTED_SIGMA: None,
        }

    def __iter__(self):
        """
        Make the instances of the class iterable objects.

        Allows to apply functions such as __dict__ to them.
        """
        yield PREDICTOR, self.predictor
        yield PREDICTION_ID, self.prediction_id
        yield PREDICTION_RESULTS, self.prediction_results
        yield PREDICTION_ARRAYS, self.prediction_arrays
        yield PREDICTION_SUMMARY, {
            i_key: dict(self.prediction_summary[i_key])
            for i_key in self.prediction_summary.keys()
        }
        yield LOWER_BOUND, self.lower_bound
        yield UPPER_BOUND, self.upper_bound

    def condense_to_arrays(self):
        """
        Add prediction result to EnsemblePrediction object.

        Reshape the prediction results to an array and add them as a
        member to the EnsemblePrediction objects. It's meant to be used only
        if all conditions of a prediction have the same observables, as this
        is often the case for large-scale data sets taken from online
        databases or similar.
        """
        # prepare for storing results over all predictions
        output = []
        output_sensi = []
        timepoints = []

        for result in self.prediction_results:
            # stack outputs, output sensitivities and timepoints to one array
            # use first element as dummy, to see if outputs have been computed
            if result.conditions[0].output is not None:
                output.append(
                    np.concatenate(
                        [cond.output for cond in result.conditions], axis=0
                    )
                )
            else:
                output = None

            if result.conditions[0].output_sensi is not None:
                output_sensi.append(
                    np.concatenate(
                        [cond.output_sensi for cond in result.conditions],
                        axis=0,
                    )
                )
            else:
                output_sensi = None

            timepoints.append(
                np.concatenate(
                    [cond.timepoints for cond in result.conditions], axis=0
                )
            )

        # stack results in third dimension
        if output is not None:
            output = np.stack(output, axis=2)
        if output_sensi is not None:
            output_sensi = np.stack(output_sensi, axis=3)

        # formulate as dict
        self.prediction_arrays = {
            OUTPUT: output,
            OUTPUT_SENSI: output_sensi,
            TIMEPOINTS: np.stack(timepoints, axis=-1),
        }

    def compute_summary(
        self,
        percentiles_list: Sequence[int] = (5, 20, 80, 95),
        weighting: bool = False,
        compute_weighted_sigma: bool = False,
    ) -> Dict:
        """
        Compute summary from the ensemble prediction results.

        Summary includes the mean, the median, the standard deviation and
        possibly percentiles. Those summary results are added as a data
        member to the EnsemblePrediction object.

        Parameters
        ----------
        percentiles_list:
            List or tuple of percent numbers for the percentiles
        weighting:
            Whether weights should be used for trajectory.
        compute_weighted_sigma:
            Whether weighted standard deviation of the ensemble mean trajectory
            should be computed. Defaults to False.

        Returns
        -------
        summary:
            dictionary of predictions results with the keys mean, std, median,
            percentiles, ...
        """
        # check if prediction results are available
        if not self.prediction_results:
            raise ArithmeticError(
                'Cannot compute summary statistics from '
                'empty prediction results.'
            )
        # if weightings shall be used, check whether weights are there
        if weighting:
            if not self.prediction_results[0].conditions[0].output_weight:
                raise ValueError(
                    'There are no weights in the ' 'prediction results.'
                )

        n_conditions = len(self.prediction_results[0].conditions)

        def _stack_outputs(ic: int):
            """
            Stack outputs.

            Group outputs for different parameter vectors of one ensemble
            together, if they belong to the same simulation condition, and
            stack them in one array.
            """
            # Were outputs computed
            if self.prediction_results[0].conditions[ic].output is None:
                return None
            # stack predictions
            output_list = [
                prediction.conditions[ic].output
                for prediction in self.prediction_results
            ]
            # stack into one numpy array
            return np.stack(output_list, axis=-1)

        def _stack_outputs_sensi(ic: int):
            """
            Stack output sensitivities.

            Group output sensitivities for different parameter vectors of one
            ensemble together, if they belong to the same simulation condition,
            and stack them in one array.
            """
            # Were output sensitivitiess computed
            if self.prediction_results[0].conditions[ic].output_sensi is None:
                return None
            # stack predictions
            output_sensi_list = [
                prediction.conditions[ic].output_sensi
                for prediction in self.prediction_results
            ]
            # stack into one numpy array
            return np.stack(output_sensi_list, axis=-1)

        def _stack_weights(ic: int) -> np.ndarray:
            """
            Stack weights.

            Group weights for different parameter vectors of one ensemble
            together, if they belong to the same simulation condition, and
            stack them in one array

            Parameters
            ----------
            ic: the condition number.

            Returns
            -------
            The stacked weights.
            """
            # Were outputs computed
            if self.prediction_results[0].conditions[ic].output_weight is None:
                return None
            # stack predictions
            output_weight_list = [
                prediction.conditions[ic].output_weight
                for prediction in self.prediction_results
            ]
            # stack into one numpy array
            return np.stack(output_weight_list, axis=-1)

        def _stack_sigmas(ic: int):
            """
            Stack sigmas.

            Group sigmas for different parameter vectors of one ensemble
            together, if they belong to the same simulation condition, and
            stack them in one array.
            """
            # Were outputs computed
            if self.prediction_results[0].conditions[ic].output_sigmay is None:
                return None
            # stack predictions
            output_sigmay_list = [
                prediction.conditions[ic].output_sigmay
                for prediction in self.prediction_results
            ]
            # stack into one numpy array
            return np.stack(output_sigmay_list, axis=-1)

        def _compute_summary(
            tmp_array, percentiles_list, weights, tmp_sigmas=None
        ):
            """
            Compute summary for a set of stacked simulations.

            Summary includes means, standard deviation, median, and requested
            percentiles.
            """
            summary = {}
            summary[MEAN] = np.average(tmp_array, axis=-1, weights=weights)
            summary[STANDARD_DEVIATION] = np.sqrt(
                np.average(
                    (tmp_array.T - summary[MEAN].T).T ** 2,
                    axis=-1,
                    weights=weights,
                )
            )
            summary[MEDIAN] = np.median(tmp_array, axis=-1)
            if tmp_sigmas is not None:
                summary[WEIGHTED_SIGMA] = np.sqrt(
                    np.average(tmp_sigmas**2, axis=-1, weights=weights)
                )
            for perc in percentiles_list:
                summary[get_percentile_label(perc)] = np.percentile(
                    tmp_array, perc, axis=-1
                )
            return summary

        # preallocate for results
        cond_lists = {MEAN: [], STANDARD_DEVIATION: [], MEDIAN: []}
        if compute_weighted_sigma:
            cond_lists[WEIGHTED_SIGMA] = []
        for perc in percentiles_list:
            cond_lists[get_percentile_label(perc)] = []

        # iterate over conditions, compute summary
        for i_cond in range(n_conditions):
            # use some short hand
            current_cond = self.prediction_results[0].conditions[i_cond]

            # create a temporary array with all the outputs needed and wanted
            tmp_output = _stack_outputs(i_cond)
            tmp_output_sensi = _stack_outputs_sensi(i_cond)
            tmp_weights = np.ones(tmp_output.shape[-1])
            if weighting:
                # take exp() to get the likelihood values,
                # as the weights would be the log likelihoods.
                tmp_weights = np.exp(_stack_weights(i_cond))
            tmp_sigmas = None
            if compute_weighted_sigma:
                tmp_sigmas = _stack_sigmas(i_cond)

            # handle outputs
            if tmp_output is not None:
                output_summary = _compute_summary(
                    tmp_output, percentiles_list, tmp_weights, tmp_sigmas
                )
            else:
                output_summary = {i_key: None for i_key in cond_lists.keys()}

            # handle output sensitivities
            if tmp_output_sensi is not None:
                output_sensi_summary = _compute_summary(
                    tmp_output_sensi, percentiles_list
                )
            else:
                output_sensi_summary = {
                    i_key: None for i_key in cond_lists.keys()
                }

            # create some PredictionConditionResult to have an easier creation
            # of PredictionResults for the summaries later on
            for i_key in cond_lists.keys():
                cond_lists[i_key].append(
                    PredictionConditionResult(
                        timepoints=current_cond.timepoints,
                        output=output_summary[i_key],
                        output_sensi=output_sensi_summary[i_key],
                        output_ids=current_cond.output_ids,
                    )
                )

        self.prediction_summary = {
            i_key: PredictionResult(
                conditions=cond_lists[i_key],
                condition_ids=self.prediction_results[0].condition_ids,
                comment=str(i_key),
            )
            for i_key in cond_lists.keys()
        }

        # also return the object
        return self.prediction_summary

    def compute_chi2(self, amici_objective: AmiciObjective):
        """
        Compute the chi^2 error of the weighted mean trajectory.

        Parameters
        ----------
        amici_objective:
            The objective function of the model,
            the parameter ensemble was created from.

        Returns
        -------
        The chi^2 error.
        """
        if (self.prediction_summary[MEAN] is None) or (
            self.prediction_summary[WEIGHTED_SIGMA] is None
        ):
            try:
                self.compute_summary(
                    weighting=True, compute_weighted_sigma=True
                )
            except TypeError:
                raise ValueError('Computing a summary failed.')
        n_conditions = len(self.prediction_results[0].conditions)
        chi_2 = []
        for i_cond in range(n_conditions):
            # get measurements and put into right form
            y_meas = amici_objective.edatas[i_cond].getObservedData()
            y_meas = np.array(y_meas)
            # bring into shape (n_t,n_y)
            y_meas = y_meas.reshape(
                amici_objective.edatas[0].nt(),
                amici_objective.edatas[0].nytrue(),
            )
            mean_traj = self.prediction_summary[MEAN].conditions[i_cond].output
            weighted_sigmas = (
                self.prediction_summary[WEIGHTED_SIGMA]
                .conditions[i_cond]
                .output
            )
            if y_meas.shape != mean_traj.shape:
                raise ValueError(
                    'Shape of trajectory and shape '
                    'of measurements does not match.'
                )
            chi_2.append(
                np.nansum(((y_meas - mean_traj) / weighted_sigmas) ** 2)
            )
        return np.sum(chi_2)


class Ensemble:
    """
    An ensemble is a wrapper around a numpy array.

    It comes with some convenience functionality: It allows to map parameter
    values via identifiers to the correct parameters, it allows to compute
    summaries of the parameter vectors (mean, standard deviation, median,
    percentiles) more easily, and it can store predictions made by pyPESTO,
    such that the parameter ensemble and the predictions are linked to each
    other.
    """

    def __init__(
        self,
        x_vectors: np.ndarray,
        x_names: Sequence[str] = None,
        vector_tags: Sequence[Tuple[int, int]] = None,
        ensemble_type: EnsembleType = None,
        predictions: Sequence[EnsemblePrediction] = None,
        lower_bound: np.ndarray = None,
        upper_bound: np.ndarray = None,
    ):
        """
        Initialize.

        Parameters
        ----------
        x_vectors:
            parameter vectors of the ensemble, in the format
            n_parameter x n_vectors
        x_names:
            Names or identifiers of the parameters
        vector_tags:
            Additional tag, which adds information about the the parameter
            vectors of the form (optimization_run, optimization_step) if the
            ensemble is created from an optimization result or
            (sampling_chain, sampling_step) if the ensemble is created from a
            sampling result.
        ensemble_type:
            Type of ensemble: Ensemble (default), sample, or unprocessed_chain
            Samples are meant to be representative, ensembles can be any
            ensemble of parameters, and unprocessed chains still have burn-ins
        predictions:
            List of EnsemblePrediction objects
        lower_bound:
            array of potential lower bounds for the parameters
        upper_bound:
            array of potential upper bounds for the parameters
        """
        # Do we have a representative sample or just random ensemble?
        self.ensemble_type = EnsembleType.ensemble
        if ensemble_type is not None:
            self.ensemble_type = ensemble_type

        # handle parameter vectors and sizes
        self.x_vectors = x_vectors
        self.n_x = x_vectors.shape[0]
        self.n_vectors = x_vectors.shape[1]
        self.vector_tags = vector_tags
        self.summary = None

        # store bounds
        self.lower_bound = np.full((self.n_x,), np.nan)
        if lower_bound is not None:
            if np.array(lower_bound).size == 1:
                self.lower_bound = np.full((x_vectors.shape[0],), lower_bound)
            else:
                self.lower_bound = lower_bound
        self.upper_bound = np.full(self.n_x, np.nan)
        if upper_bound is not None:
            if np.array(upper_bound).size == 1:
                self.upper_bound = np.full(x_vectors.shape[0], upper_bound)
            else:
                self.upper_bound = upper_bound

        # handle parameter names
        if x_names is not None:
            self.x_names = x_names
        else:
            self.x_names = [f'x_{ix}' for ix in range(self.n_x)]

        # Do we have predictions for this ensemble?
        self.predictions = []
        if predictions is not None:
            self.predictions = predictions

    @staticmethod
    def from_sample(
        result: Result,
        remove_burn_in: bool = True,
        chain_slice: slice = None,
        x_names: Sequence[str] = None,
        lower_bound: np.ndarray = None,
        upper_bound: np.ndarray = None,
        **kwargs,
    ):
        """
        Construct an ensemble from a sample.

        Parameters
        ----------
        result:
            A pyPESTO result that contains a sample result.
        remove_burn_in:
            Exclude parameter vectors from the ensemble if they are in the
            "burn-in".
        chain_slice:
            Subset the chain with a slice. Any "burn-in" removal occurs first.
        x_names:
            Names or identifiers of the parameters
        lower_bound:
            array of potential lower bounds for the parameters
        upper_bound:
            array of potential upper bounds for the parameters

        Returns
        -------
        The ensemble.
        """
        x_vectors = result.sample_result.trace_x[0]
        if x_names is None:
            x_names = [
                result.problem.x_names[i]
                for i in result.problem.x_free_indices
            ]
        if lower_bound is None:
            lower_bound = result.problem.lb
        if upper_bound is None:
            upper_bound = result.problem.ub
        if remove_burn_in:
            if result.sample_result.burn_in is None:
                geweke_test(result)
            burn_in = result.sample_result.burn_in
            x_vectors = x_vectors[burn_in:]
        if chain_slice is not None:
            x_vectors = x_vectors[chain_slice]
        x_vectors = x_vectors.T
        return Ensemble(
            x_vectors=x_vectors,
            x_names=x_names,
            lower_bound=lower_bound,
            upper_bound=upper_bound,
            **kwargs,
        )

    @staticmethod
    def from_optimization_endpoints(
        result: Result,
        rel_cutoff: float = None,
        max_size: int = np.inf,
        percentile: float = None,
        **kwargs,
    ):
        """
        Construct an ensemble from an optimization result.

        Parameters
        ----------
        result:
            A pyPESTO result that contains an optimization result.
        rel_cutoff:
            Relative cutoff. Exclude parameter vectors, for which the
            objective value difference to the best vector is greater than
            cutoff, i.e. include all vectors such that
            `fval(vector) <= fval(opt_vector) + rel_cutoff`.
        max_size:
            The maximum size the ensemble should be.
        percentile:
            Percentile of a chi^2 distribution. Used to determine the
            cutoff value.

        Returns
        -------
        The ensemble.
        """
        if rel_cutoff is None and percentile is None:
            abs_cutoff = np.inf
        elif rel_cutoff is not None:
            abs_cutoff = result.optimize_result[0].fval + rel_cutoff
            if percentile is not None:
                logger.warning(
                    'percentile is going to be ignored as '
                    'rel_cutoff is not `None`.'
                )
        else:
            abs_cutoff = calculate_cutoff(result=result, percentile=percentile)
        x_vectors = []
        vector_tags = []
        x_names = [
            result.problem.x_names[i] for i in result.problem.x_free_indices
        ]

        for start in result.optimize_result.list:
            # add the parameters from the next start as long as we
            # did not reach maximum size and the next value is still
            # lower than the cutoff value
            if start['fval'] <= abs_cutoff and len(x_vectors) < max_size:
                x_vectors.append(start['x'])

                # the vector tag will be a -1 to indicate it is the last step
                vector_tags.append((int(start['id']), -1))
            else:
                break

        # print a warning if there are no vectors within the ensemble
        if len(x_vectors) == 0:
            raise ValueError(
                'The ensemble does not contain any vectors. '
                'Either the cutoff value was too small\n or the '
                'result.optimize_result object might be empty.'
            )
        elif len(x_vectors) < max_size:
            logger.info(
                f'The ensemble contains {len(x_vectors)} parameter '
                'vectors, which is less than the maximum size.\nIf '
                'you want to include more \nvectors, you can consider '
                'raising the cutoff value or including parameters '
                'from \nthe history with the `from_history` function.'
            )

        x_vectors = np.stack(x_vectors, axis=1)
        return Ensemble(
            x_vectors=x_vectors,
            x_names=x_names,
            vector_tags=vector_tags,
            lower_bound=result.problem.lb,
            upper_bound=result.problem.ub,
            **kwargs,
        )

    @staticmethod
    def from_optimization_history(
        result: Result,
        rel_cutoff: float = None,
        max_size: int = np.inf,
        max_per_start: int = np.inf,
        distribute: bool = True,
        percentile: float = None,
        **kwargs,
    ):
        """
        Construct an ensemble from the history of an optimization.

        Parameters
        ----------
        result:
            A pyPESTO result that contains an optimization result
            with history recorded.
        rel_cutoff:
            Relative cutoff. Exclude parameter vectors, for which the
            objective value difference to the best vector is greater than
            cutoff, i.e. include all vectors such that
            `fval(vector) <= fval(opt_vector) + rel_cutoff`.
        max_size:
            The maximum size the ensemble should be.
        max_per_start:
            The maximum number of vectors to be included from a
            single optimization start.
        distribute:
            Boolean flag, whether the best (False) values from the
            start should be taken or whether the indices should be
            more evenly distributed.
        percentile:
            Percentile of a chi^2 distribution. Used to determinie the
            cutoff value.

        Returns
        -------
        The ensemble.
        """
        if rel_cutoff is None and percentile is None:
            abs_cutoff = np.inf
        elif rel_cutoff is not None:
            abs_cutoff = result.optimize_result[0].fval + rel_cutoff
        else:
            abs_cutoff = calculate_cutoff(result=result, percentile=percentile)
        if not result.optimize_result.list[0].history.options['trace_record']:
            logger.warning(
                'The optimize result has no trace. The Ensemble '
                'will automatically be created through '
                'from_optimization_endpoints().'
            )
            return Ensemble.from_optimization_endpoints(
                result=result,
                rel_cutoff=rel_cutoff,
                max_size=max_size,
                **kwargs,
            )
        x_vectors = []
        vector_tags = []
        x_names = [
            result.problem.x_names[i] for i in result.problem.x_free_indices
        ]
        lb = result.problem.lb
        ub = result.problem.ub

        # calculate the number of starts whose final nllh is below cutoff
        n_starts = sum(
            start['fval'] <= abs_cutoff
            for start in result.optimize_result.list
        )

        fval_trace = [
            np.array(
                result.optimize_result.list[i_ms][HISTORY].get_fval_trace()
            )
            for i_ms in range(n_starts)
        ]
        x_trace = [
            result.optimize_result.list[i_ms][HISTORY].get_x_trace()
            for i_ms in range(n_starts)
        ]

        # calculate the number of iterations included from each start
        n_per_starts = entries_per_start(
            fval_traces=fval_trace,
            cutoff=abs_cutoff,
            max_per_start=max_per_start,
            max_size=max_size,
        )
        # determine x_vectors from each start
        for start in range(n_starts):
            indices = get_vector_indices(
                trace_start=fval_trace[start],
                cutoff=abs_cutoff,
                n_vectors=n_per_starts[start],
                distribute=distribute,
            )
            x_vectors.extend([x_trace[start][ind] for ind in indices])
            vector_tags.extend(
                [
                    (int(result.optimize_result.list[start]['id']), ind)
                    for ind in indices
                ]
            )

        # raise a `ValueError` if there are no vectors within the ensemble
        if len(x_vectors) == 0:
            raise ValueError(
                'The ensemble does not contain any vectors. '
                'Either the `cutoff` value was too \nsmall '
                'or the `result.optimize_result` object might '
                'be empty.'
            )

        x_vectors = np.stack(x_vectors, axis=1)
        return Ensemble(
            x_vectors=x_vectors,
            x_names=x_names,
            vector_tags=vector_tags,
            lower_bound=lb,
            upper_bound=ub,
            **kwargs,
        )

    def __iter__(self):
        """
        Make the instances of the class iterable objects.

        Allows to apply functions such as __dict__ to them.
        """
        yield X_VECTOR, self.x_vectors
        yield NX, self.n_x
        yield X_NAMES, self.x_names
        yield NVECTORS, self.n_vectors
        yield VECTOR_TAGS, self.vector_tags
        yield ENSEMBLE_TYPE, self.ensemble_type
        yield PREDICTIONS, self.predictions
        yield SUMMARY, self.summary
        yield LOWER_BOUND, self.lower_bound
        yield UPPER_BOUND, self.upper_bound

    def _map_parameters_by_objective(
        self,
        predictor: Callable,
        default_value: float = None,
    ):
        """
        Create mapping for parameters from ensebmle to predictor.

        The parameters of the ensemble don't need to have the same ordering as
        in the predictor.
        """
        # create short hands
        parameter_ids_objective = predictor.amici_objective.x_names
        parameter_ids_ensemble = self.x_names
        # map, and fill with `default_value` if not found and `default_value`
        # is specified.
        mapping = []
        for parameter_id_objective in parameter_ids_objective:
            if parameter_id_objective in parameter_ids_ensemble:
                mapping.append(
                    parameter_ids_ensemble.index(parameter_id_objective)
                )
            elif default_value is not None:
                mapping.append(default_value)
        return mapping

    def predict(
        self,
        predictor: Callable,
        prediction_id: str = None,
        sensi_orders: Tuple = (0,),
        default_value: float = None,
        mode: ModeType = MODE_FUN,
        include_llh_weights: bool = False,
        include_sigmay: bool = False,
        engine: Engine = None,
        progress_bar: bool = True,
    ) -> EnsemblePrediction:
        """
        Run predictions for a full ensemble.

        User needs to hand over a predictor function and settings, then all
        results are grouped as EnsemblePrediction for the whole ensemble

        Parameters
        ----------
        predictor:
            Prediction function, e.g., an AmiciPredictor
        prediction_id:
            Identifier for the predictions
        sensi_orders:
            Specifies which sensitivities to compute, e.g. (0,1) -> fval, grad
        default_value:
            If parameters are needed in the mapping, which are not found in the
            parameter source, it can make sense to fill them up with this
            default value (e.g. `np.nan`) in some cases (to be used with
            caution though).
        mode:
            Whether to compute function values or residuals.
        include_llh_weights:
            Whether to include weights in the output of the predictor.
        include_sigmay:
            Whether to include standard deviations in the output
            of the predictor.
        engine:
            Parallelization engine. Defaults to sequential execution on a
            `SingleCoreEngine`.
        progress_bar:
            Whether to display a progress bar.

        Returns
        -------
        The prediction of the ensemble.
        """
        if engine is None:
            engine = SingleCoreEngine()

        # Vectors are chunked to improve parallization performance.
        n_chunks = self.n_vectors  # Default is no chunking.
        if isinstance(engine, MultiProcessEngine):
            n_chunks = engine.n_procs
        if isinstance(engine, MultiThreadEngine):
            n_chunks = engine.n_threads
        chunks = [
            (
                (chunk_i + 0) * int(np.floor(self.n_vectors / n_chunks)),
                (chunk_i + 1) * int(np.floor(self.n_vectors / n_chunks)),
            )
            for chunk_i in range(n_chunks)
        ]
        # Last chunk should contain any remaining vectors that may have
        # been skipped due to the `floor` method.
        chunks[-1] = (chunks[-1][0], self.n_vectors)

        # Get the correct parameter mapping.
        mapping = self._map_parameters_by_objective(
            predictor,
            default_value=default_value,
        )

        # Setup the tasks with the prediction method and chunked vectors.
        method = partial(
            predictor,
            sensi_orders=sensi_orders,
            mode=mode,
            include_sigmay=include_sigmay,
            include_llh_weights=include_llh_weights,
        )
        tasks = [
            EnsembleTask(
                method=method,
                vectors=self.x_vectors[mapping, chunk_start:chunk_end],
                id=chunk_i,
            )
            for chunk_i, (chunk_start, chunk_end) in enumerate(chunks)
        ]

        # Execute tasks and flatten chunked results.
        prediction_results = [
            prediction_result
            for prediction_chunk in engine.execute(
                tasks, progress_bar=progress_bar
            )
            for prediction_result in prediction_chunk
        ]

        return EnsemblePrediction(
            predictor=predictor,
            prediction_id=prediction_id,
            prediction_results=prediction_results,
        )

    def compute_summary(
        self, percentiles_list: Sequence[int] = (5, 20, 80, 95)
    ):
        """
        Compute summary for the parameters of the ensemble.

        Summary includes the mean, the median, the standard deviation and
        possibly percentiles. Those summary results are added as a data
        member to the EnsemblePrediction object.

        Parameters
        ----------
        percentiles_list:
            List or tuple of percent numbers for the percentiles

        Returns
        -------
        summary:
            Dict with mean, std, median, and percentiles of parameter vectors
        """
        # compute summaries based on parameters
        summary = {
            MEAN: np.mean(self.x_vectors, axis=1),
            STANDARD_DEVIATION: np.std(self.x_vectors, axis=1),
            MEDIAN: np.median(self.x_vectors, axis=1),
        }
        for perc in percentiles_list:
            summary[get_percentile_label(perc)] = np.percentile(
                self.x_vectors, perc, axis=1
            )
        # store and return results
        self.summary = summary
        return summary

    def check_identifiability(self) -> pd.DataFrame:
        """
        Check identifiability of ensemble.

        Use ensemble mean and standard deviation to assess (in a rudimentary
        way) whether or not parameters are identifiable. Returns a dataframe
        with tuples, which specify whether or not the lower and the upper
        bounds are violated.

        Returns
        -------
        parameter_identifiability:
            DataFrame indicating parameter identifiability based on mean
            plus/minus standard deviations and parameter bounds
        """
        # Recompute the summary, maybe the ensemble objects has been changed.
        self.compute_summary()

        # check identifiability for each parameter
        parameter_identifiability = []
        for ix, x_name in enumerate(self.x_names):
            # define some short hands
            lb = self.lower_bound[ix]
            ub = self.upper_bound[ix]
            mean = self.summary[MEAN][ix]
            std = self.summary[STANDARD_DEVIATION][ix]
            median = self.summary[MEAN][ix]
            perc_list = [
                int(i_key[11:])
                for i_key in self.summary.keys()
                if i_key[0:4] == 'perc'
            ]
            perc_lower = [perc for perc in perc_list if perc < 50]
            perc_upper = [perc for perc in perc_list if perc > 50]

            # create dict of identifiability
            tmp_identifiability = {
                'parameterId': x_name,
                'lowerBound': lb,
                'upperBound': ub,
                'ensemble_mean': mean,
                'ensemble_std': std,
                'ensemble_median': median,
                'within lb: 1 std': lb < mean - std,
                'within ub: 1 std': ub > mean + std,
                'within lb: 2 std': lb < mean - 2 * std,
                'within ub: 2 std': ub > mean + 2 * std,
                'within lb: 3 std': lb < mean - 3 * std,
                'within ub: 3 std': ub > mean + 3 * std,
            }
            # handle percentiles
            for perc in perc_lower:
                tmp_identifiability[f'within lb: perc {perc}'] = (
                    lb < self.summary[get_percentile_label(perc)][ix]
                )
            for perc in perc_upper:
                tmp_identifiability[f'within ub: perc {perc}'] = (
                    ub > self.summary[get_percentile_label(perc)][ix]
                )

            parameter_identifiability.append(tmp_identifiability)

        # create DataFrame
        parameter_identifiability = pd.DataFrame(parameter_identifiability)
        parameter_identifiability.index = parameter_identifiability[
            'parameterId'
        ]

        return parameter_identifiability


def entries_per_start(
    fval_traces: List['np.ndarray'],
    cutoff: float,
    max_size: int,
    max_per_start: int,
):
    """
    Create the indices of each start that will be included in the ensemble.

    Parameters
    ----------
    fval_traces:
        the fval-trace of each start.
    cutoff:
        Exclude parameters from the optimization if the nllh
        is higher than the `cutoff`.
    max_size:
        The maximum size the ensemble should be.
    max_per_start:
        The maximum number of vectors to be included from a
        single optimization start.

    Returns
    -------
        A list of number of candidates per start that are to
        be included in the ensemble.
    """
    # choose possible candidates
    ens_ind = [np.flatnonzero(fval <= cutoff) for fval in fval_traces]

    # count the number of candidates per start
    n_per_start = np.array([len(start) for start in ens_ind])

    # if all possible indices can be included, return
    if (n_per_start < max_per_start).all() and sum(n_per_start) < max_size:
        return n_per_start

    # trimm down starts that exceed the limit:
    n_per_start = [min(n, max_per_start) for n in n_per_start]

    # trimm down more until it fits the max size
    decr = 0
    while sum(n_per_start) > max_size:
        n_per_start = [min(n, max_per_start - decr) for n in n_per_start]
        decr += 1
    # TODO: Possibility. With this implementation we could
    #  in a scenario, where we have more candidates than
    #  max size end up with an ensemble of size
    #  `max_size - len(n_starts)` in the worst case. We could introduce
    #  a flag which would be `force_max`, that indicates
    #  whether those remaining free slots should be filled by
    #  entries from certain starts. This would brng up the
    #  discussion which starts to choose. One obvious choice
    #  would be the best starts based on their endpoint.

    return n_per_start


def get_vector_indices(
    trace_start: np.ndarray,
    cutoff: float,
    n_vectors: int,
    distribute: bool,
):
    """
    Return the indices to be taken into an ensemble.

    Parameters
    ----------
    trace_start:
        The fval_trace of a single start.
    cutoff:
        Exclude parameters from the optimization if the nllh
        is higher than the `cutoff`.
    n_vectors:
        The number of indices to be included from one start.
    distribute:
        Boolean flag, whether the best (False) values from the
        start should be taken or whether the indices should be
        more evenly distributed.

    Returns
    -------
        The indices to include in the ensemble.
    """
    candidates = np.flatnonzero(trace_start <= cutoff)

    if distribute:
        indices = np.round(np.linspace(0, len(candidates) - 1, n_vectors))
        return candidates[indices.astype(int)]
    else:
        return candidates[:n_vectors]


def get_percentile_label(percentile: Union[float, int, str]) -> str:
    """Convert a percentile to a label.

    Labels for percentiles are used at different locations (e.g. ensemble
    prediction code, and visualization code). This method ensures that the same
    percentile is labeled identically everywhere.

    The percentile is rounded to two decimal places in the label representation
    if it is specified to more decimal places. This is for readability in
    plotting routines, and to avoid float to string conversion issues related
    to float precision.

    Parameters
    ----------
    percentile:
        The percentile value that will be used to generate a label.

    Returns
    -------
    The label of the (possibly rounded) percentile.
    """
    if isinstance(percentile, str):
        percentile = float(percentile)
        if percentile == round(percentile):
            percentile = round(percentile)
    if isinstance(percentile, float):
        percentile_str = f'{percentile:.2f}'
        # Add `...` to the label if the percentile value changed after rounding
        if float(percentile_str) != percentile:
            percentile_str += '...'
        percentile = percentile_str
    return f'{PERCENTILE} {percentile}'


def calculate_cutoff(result: Result, percentile: float = 0.95):
    """
    Calculate the cutoff of the ensemble.

    Based on the number of parameters of the problem. Based on the
    assumption that the difference of the nllh's of the true and optimal
    parameter is chi^2 distributed with n_theta degress of freedom.

    Parameters
    ----------
    result:
        The optimization result from which to create the ensemble.
    percentile:
        The percentile of the chi^2 distribution. Between 0 and 100.
        Higher values will result in a more lax cutoff. If the value is greater
        than 100, the cutoff will be returned as np.inf.

    Returns
    -------
    The calculated cutoff value.
    """
    if percentile > 100:
        raise ValueError(
            f"percentile={percentile} is too large. Choose "
            f"0<=percentile<=100."
        )
    # optimal point as base:
    fval_opt = result.optimize_result[0].fval
    # degrees of freedom is equal to the number of parameters
    df = result.problem.dim
    range = chi2.ppf(q=percentile / 100, df=df)

    return fval_opt + range



"""
Aesara models interface.

Adds an interface for the construction of loss functions
incorporating aesara models. This permits computation of derivatives using a
combination of objective based methods and aesara based backpropagation.
"""

import copy
from typing import Optional, Sequence, Tuple

import numpy as np

from ...C import FVAL, GRAD, HESS, MODE_FUN, RDATAS, ModeType
from ..base import ObjectiveBase, ResultDict

try:
    import aesara
    import aesara.tensor as aet
    from aesara.tensor import Op
    from aesara.tensor.var import TensorVariable
except ImportError:
    raise ImportError(
        "Using an aeasara objective requires an installation of "
        "the python package aesara. Please install aesara via "
        "`pip install aesara`."
    )


class AesaraObjective(ObjectiveBase):
    """
    Wrapper around an ObjectiveBase.

    Computes the gradient at each evaluation, caching it for later calls.
    Caching is only enabled after the first time the gradient is asked for
    and disabled whenever the cached gradient is not used, in order not to
    increase computation time for derivative-free samplers.

    Parameters
    ----------
    objective:
        The `pypesto.ObjectiveBase` to wrap.
    aet_x:
        Tensor variables that define the variables of `aet_fun`
    aet_fun:
        Aesara function that maps `aet_x` to the variables of `objective`
    coeff:
        Multiplicative coefficient for objective
    """

    def __init__(
        self,
        objective: ObjectiveBase,
        aet_x: TensorVariable,
        aet_fun: TensorVariable,
        coeff: Optional[float] = 1.0,
        x_names: Sequence[str] = None,
    ):
        if not isinstance(objective, ObjectiveBase):
            raise TypeError('objective must be an ObjectiveBase instance')
        if not objective.check_mode(MODE_FUN):
            raise NotImplementedError(
                f'objective must support mode={MODE_FUN}'
            )
        super().__init__(x_names)
        self.base_objective = objective

        self.aet_x = aet_x
        self.aet_fun = aet_fun
        self._coeff = coeff

        self.obj_op = AesaraObjectiveOp(self, self._coeff)

        # compiled function
        if objective.has_fun:
            self.afun = aesara.function([aet_x], self.obj_op(aet_fun))

        # compiled gradient
        if objective.has_grad:
            self.agrad = aesara.function(
                [aet_x], aesara.grad(self.obj_op(aet_fun), [aet_x])
            )

        # compiled hessian
        if objective.has_hess:
            self.ahess = aesara.function(
                [aet_x], aesara.gradient.hessian(self.obj_op(aet_fun), [aet_x])
            )

        # compiled input mapping
        self.infun = aesara.function([aet_x], aet_fun)

        # temporary storage for evaluation results of objective
        self.cached_base_ret: ResultDict = {}

    def check_mode(self, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        return mode == MODE_FUN

    def check_sensi_orders(self, sensi_orders, mode: ModeType) -> bool:
        """See `ObjectiveBase` documentation."""
        if not self.check_mode(mode):
            return False
        else:
            return self.base_objective.check_sensi_orders(sensi_orders, mode)

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        **kwargs,
    ) -> ResultDict:
        """
        See `ObjectiveBase` for more documentation.

        Main method to overwrite from the base class. It handles and
        delegates the actual objective evaluation.
        """
        # hess computation in aesara requires grad
        if 2 in sensi_orders and 1 not in sensi_orders:
            sensi_orders = (1, *sensi_orders)

        # this computes all the results from the inner objective, rendering
        # them accessible to aesara compiled functions

        set_return_dict, return_dict = (
            'return_dict' in kwargs,
            kwargs.pop('return_dict', False),
        )
        self.cached_base_ret = self.base_objective(
            self.infun(x), sensi_orders, mode, return_dict=True, **kwargs
        )
        if set_return_dict:
            kwargs['return_dict'] = return_dict
        ret = {}
        if RDATAS in self.cached_base_ret:
            ret[RDATAS] = self.cached_base_ret[RDATAS]
        if 0 in sensi_orders:
            ret[FVAL] = float(self.afun(x))
        if 1 in sensi_orders:
            ret[GRAD] = self.agrad(x)[0]
        if 2 in sensi_orders:
            ret[HESS] = self.ahess(x)[0]

        return ret

    def __deepcopy__(self, memodict=None):
        other = AesaraObjective(
            copy.deepcopy(self.base_objective),
            self.aet_x,
            self.aet_fun,
            self._coeff,
        )

        return other


class AesaraObjectiveOp(Op):
    """
    Aesara wrapper around a (non-normalized) log-probability function.

    Parameters
    ----------
    obj:
        Base aesara objective
    coeff:
        Multiplicative coefficient for the objective function value
    """

    itypes = [aet.dvector]  # expects a vector of parameter values when called
    otypes = [aet.dscalar]  # outputs a single scalar value (the log prob)

    def __init__(self, obj: AesaraObjective, coeff: Optional[float] = 1.0):
        self._objective: AesaraObjective = obj
        self._coeff: float = coeff

        # initialize the sensitivity Op
        if obj.has_grad:
            self._log_prob_grad = AesaraObjectiveGradOp(obj, coeff)
        else:
            self._log_prob_grad = None

    def perform(self, node, inputs, outputs, params=None):  # noqa
        # note that we use precomputed values from the outer
        # AesaraObjective.call_unprocessed here, which means we can
        # ignore inputs here
        log_prob = self._coeff * self._objective.cached_base_ret[FVAL]
        outputs[0][0] = np.array(log_prob)

    def grad(self, inputs, g):
        """
        Calculate the hessian.

        Actually returns the vector-hessian product - g[0] is a vector of
        parameter values.
        """
        if self._log_prob_grad is None:
            return super(AesaraObjectiveOp, self).grad(inputs, g)
        (theta,) = inputs
        log_prob_grad = self._log_prob_grad(theta)
        return [g[0] * log_prob_grad]


class AesaraObjectiveGradOp(Op):
    """
    Aesara wrapper around a (non-normalized) log-probability gradient function.

    This Op will be called with a vector of values and also return a vector of
    values - the gradients in each dimension.

    Parameters
    ----------
    obj:
        Base aesara objective
    coeff:
        Multiplicative coefficient for the objective function value
    """

    itypes = [aet.dvector]  # expects a vector of parameter values when called
    otypes = [aet.dvector]  # outputs a vector (the log prob grad)

    def __init__(self, obj: AesaraObjective, coeff: Optional[float] = 1.0):
        self._objective: AesaraObjective = obj
        self._coeff: float = coeff

        if obj.has_hess:
            self._log_prob_hess = AesaraObjectiveHessOp(obj, coeff)
        else:
            self._log_prob_hess = None

    def perform(self, node, inputs, outputs, params=None):  # noqa
        # note that we use precomputed values from the outer
        # AesaraObjective.call_unprocessed here, which means we can
        # ignore inputs here
        log_prob_grad = self._coeff * self._objective.cached_base_ret[GRAD]
        outputs[0][0] = log_prob_grad

    def grad(self, inputs, g):
        """
        Calculate the hessian.

        Actually returns the vector-hessian product - g[0] is a vector of
        parameter values.
        """
        if self._log_prob_hess is None:
            return super(AesaraObjectiveGradOp, self).grad(inputs, g)
        (theta,) = inputs
        log_prob_hess = self._log_prob_hess(theta)
        return [g[0].dot(log_prob_hess)]


class AesaraObjectiveHessOp(Op):
    """
    Aesara wrapper around a (non-normalized) log-probability Hessian function.

    This Op will be called with a vector of values and also return a matrix of
    values - the Hessian in each dimension.

    Parameters
    ----------
    obj:
        Base aesara objective
    coeff:
        Multiplicative coefficient for the objective function value
    """

    itypes = [aet.dvector]
    otypes = [aet.dmatrix]

    def __init__(self, obj: AesaraObjective, coeff: Optional[float] = 1.0):
        self._objective: AesaraObjective = obj
        self._coeff: float = coeff

    def perform(self, node, inputs, outputs, params=None):  # noqa
        # note that we use precomputed values from the outer
        # AesaraObjective.call_unprocessed here, which means we can
        # ignore inputs here
        log_prob_hess = self._coeff * self._objective.cached_base_ret[HESS]
        outputs[0][0] = log_prob_hess



# noqa: D400,D205
"""
Visualize
=========

pypesto comes with various visualization routines. To use these,
import pypesto.visualize.
"""

from pypesto.util import assign_clusters, delete_nan_inf

from .clust_color import assign_clustered_colors, assign_colors
from .dimension_reduction import (
    ensemble_crosstab_scatter_lowlevel,
    ensemble_scatter_lowlevel,
    projection_scatter_pca,
    projection_scatter_umap,
    projection_scatter_umap_original,
)
from .ensemble import ensemble_identifiability
from .misc import process_offset_y, process_result_list, process_y_limits
from .optimization_stats import (
    optimization_run_properties_one_plot,
    optimization_run_properties_per_multistart,
    optimization_run_property_per_multistart,
)
from .optimizer_convergence import optimizer_convergence
from .optimizer_history import optimizer_history, optimizer_history_lowlevel
from .parameters import parameter_hist, parameters, parameters_lowlevel
from .profile_cis import profile_cis
from .profiles import profile_lowlevel, profiles, profiles_lowlevel
from .reference_points import ReferencePoint, create_references
from .sampling import (
    sampling_1d_marginals,
    sampling_fval_traces,
    sampling_parameter_cis,
    sampling_parameter_traces,
    sampling_prediction_trajectories,
    sampling_scatter,
)
from .waterfall import waterfall, waterfall_lowlevel



"""
Visualization of the model fit after optimization.

Currently only for PEtab problems.
"""
from typing import Sequence, Union

import amici
import amici.plotting
import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np
import petab
from amici.petab_objective import rdatas_to_simulation_df
from petab.visualize import plot_problem

from ..C import RDATAS
from ..problem import Problem
from ..result import Result

AmiciModel = Union['amici.Model', 'amici.ModelPtr']

__all__ = ["visualize_optimized_model_fit", "time_trajectory_model"]


def visualize_optimized_model_fit(
    petab_problem: petab.Problem,
    result: Union[Result, Sequence[Result]],
    pypesto_problem: Problem,
    start_index: int = 0,
    return_dict: bool = False,
    unflattened_petab_problem: petab.Problem = None,
    **kwargs,
) -> Union[matplotlib.axes.Axes, None]:
    """
    Visualize the optimized model fit of a PEtab problem.

    Function calls the PEtab visualization file of the ``petab_problem`` and
    visualizes the fit of the optimized parameter. Common additional
    argument is ``subplot_dir`` to specify the directory each subplot is
    saved to. Further keyword arguments are delegated to
    :func:`petab.visualize.plot_with_vis_spec`, see there for more information.

    Parameters
    ----------
    petab_problem:
        The :py:class:`petab.Problem` that was optimized.
    result:
        The result object from optimization.
    start_index:
        The index of the optimization run in `result.optimize_result.list`.
        Ignored if `problem_parameters` is provided.
    pypesto_problem:
        The pyPESTO problem.
    return_dict:
        Return plot and simulation results as a dictionary.
    unflattened_petab_problem:
        If the original PEtab problem is flattened, this can be passed
        to plot with the original unflattened problem.
    kwargs:
        Passed to :func:`petab.visualize.plot_problem`.

    Returns
    -------
    axes: `matplotlib.axes.Axes` object of the created plot.
    None: In case subplots are saved to file
    """
    x = result.optimize_result.list[start_index]['x'][
        pypesto_problem.x_free_indices
    ]
    objective_result = pypesto_problem.objective(x, return_dict=True)

    simulation_df = rdatas_to_simulation_df(
        objective_result[RDATAS],
        pypesto_problem.objective.amici_model,
        petab_problem.measurement_df,
    )

    # handle flattened PEtab problems
    petab_problem_to_plot = petab_problem
    if unflattened_petab_problem:
        simulation_df = petab.core.unflatten_simulation_df(
            simulation_df=simulation_df,
            petab_problem=unflattened_petab_problem,
        )
        petab_problem_to_plot = unflattened_petab_problem

    # plot
    axes = plot_problem(
        petab_problem=petab_problem_to_plot,
        simulations_df=simulation_df,
        **kwargs,
    )
    if return_dict:
        return {
            'axes': axes,
            'objective_result': objective_result,
            'simulation_df': simulation_df,
        }
    return axes


def time_trajectory_model(
    result: Union[Result, Sequence[Result]],
    problem: Problem = None,
    # TODO: conditions: Union[str, Sequence[str]] = None,
    timepoints: Union[np.ndarray, Sequence[np.ndarray]] = None,
    n_timepoints: int = 1000,
    start_index: int = 0,
    state_ids: Union[str, Sequence[str]] = None,
    state_names: Union[str, Sequence[str]] = None,
    observable_ids: Union[str, Sequence[str]] = None,
) -> Union[matplotlib.axes.Axes, None]:
    """
    Visualize the time trajectory of the model with given timepoints.

    It does this by calling the amici plotting routines.

    Parameters
    ----------
    result:
        The result object from optimization.
    problem:
        A pypesto problem. Default is 'None' in which case result.problem is
        used. Needed in case the result is loaded from hdf5.
    timepoints:
        Array of timepoints, at which the trajectory will be plotted.
    n_timepoints:
        Number of timepoints to be plotted between 0 and last measurement of
        the model. Only used when timepoints==None.
    start_index:
        Index of Optimization run to be plotted. Default is best start.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    if problem is None:
        problem = result.problem
    # add timepoints as needed
    if timepoints is None:
        end_time = max(problem.objective.edatas[0].getTimepoints())
        timepoints = np.linspace(start=0, stop=end_time, num=n_timepoints)
    obj = problem.objective.set_custom_timepoints(timepoints_global=timepoints)

    # evaluate objective with return dic = True to get data
    parameters = result.optimize_result.list[start_index]['x']
    # reduce vector to only include free indices. Needed downstream.
    parameters = problem.get_reduced_vector(parameters)
    ret = obj(parameters, mode='mode_fun', sensi_orders=(0,), return_dict=True)

    if state_ids == [] and state_names == []:
        axes = _time_trajectory_model_without_states(
            model=problem.objective.amici_model,
            rdatas=ret['rdatas'],
            observable_ids=observable_ids,
        )
    else:
        axes = _time_trajectory_model_with_states(
            model=problem.objective.amici_model,
            rdatas=ret['rdatas'],
            state_ids=state_ids,
            state_names=state_names,
            observable_ids=observable_ids,
        )

    return axes


def _time_trajectory_model_with_states(
    model: AmiciModel,
    rdatas: Union['amici.ReturnData', Sequence['amici.ReturnData']],
    state_ids: Sequence[str],
    state_names: Sequence[str],
    observable_ids: Union[str, Sequence[str]],
):
    """
    Visualizes both, states and observables.

    Helper function for time_trajectory_model.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    state_ids:
        Ids of the states to be plotted.
    state_names:
        Names of the states to be plotted.
    observable_ids:
        Ids of the observable Ids to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    # if state_name, state_id or observable_id is not None, get indices
    # for these the AMICI plotting functions default to all indices if
    # `None` is specified.
    state_indices_by_id = []
    state_indices_by_name = []
    if state_ids is not None:
        state_indices_by_id = [
            model.getStateIds().index(state_id) for state_id in state_ids
        ]
    if state_names is not None:
        state_indices_by_name = [
            model.getStateNames().index(state_name)
            for state_name in state_names
        ]
    state_indices = list(set(state_indices_by_id + state_indices_by_name))
    if state_indices == []:
        state_indices = None

    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.getObservableIds().index(obs_id) for obs_id in observable_ids
        ]

    fig, axes = plt.subplots(len(rdatas), 2)
    # enforce two dimensions in case there is only one condition
    axes = np.atleast_2d(axes)

    for i_cond, rdata in enumerate(rdatas):
        amici.plotting.plotStateTrajectories(
            rdata=rdata,
            state_indices=state_indices,
            ax=axes[i_cond, 0],
            model=model,
        )
        amici.plotting.plotObservableTrajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond, 1],
            model=model,
        )
    return axes


def _time_trajectory_model_without_states(
    model: AmiciModel,
    rdatas: Union['amici.ReturnData', Sequence['amici.ReturnData']],
    observable_ids: Union[str, Sequence[str]],
):
    """
    Visualize both, states and observables.

    Helper function for time_trajectory_model.

    Parameters
    ----------
    model:
        The amici.Model from the model of interest. Used to annotate the plot.
    rdatas:
        The data to be plotted. Each entry in the Sequence corresponds to a
        condition.
    observable_ids:
        Ids of the observables to be plotted.

    Returns
    -------
    axes:
        `matplotlib.axes.Axes` object of the plot.
    """
    # if observable_id's is not None, get indices for these
    # the AMICI plotting functions default to all indices if `None` is
    # specified.
    observable_indices = None
    if observable_ids is not None:
        observable_indices = [
            model.getObservableIds().index(obs_id) for obs_id in observable_ids
        ]

    fig, axes = plt.subplots(len(rdatas))

    for i_cond, rdata in enumerate(rdatas):
        amici.plotting.plotObservableTrajectories(
            rdata=rdata,
            observable_indices=observable_indices,
            ax=axes[i_cond],
            model=model,
        )
    return axes



"""
Engines
=======

The execution of the multistarts can be parallelized in different ways, e.g.
multi-threaded or cluster-based. Note that it is not checked whether a single
task itself is internally parallelized.
"""

from .base import Engine
from .multi_process import MultiProcessEngine
from .multi_thread import MultiThreadEngine
from .single_core import SingleCoreEngine
from .task import Task



"""Engines with multi-threading parallelization."""
import copy
import logging
import os
from concurrent.futures import ThreadPoolExecutor
from typing import List

from tqdm import tqdm

from .base import Engine
from .task import Task

logger = logging.getLogger(__name__)


def work(task):
    """Just execute task."""
    return task.execute()


class MultiThreadEngine(Engine):
    """
    Parallelize the task execution using multithreading.

    Parameters
    ----------
    n_threads:
        The maximum number of threads to use in parallel.
        Defaults to the number of CPUs available on the system according to
        `os.cpu_count()`.
        The effectively used number of threads will be the minimum of
        `n_threads` and the number of tasks submitted.
    """

    def __init__(self, n_threads: int = None):
        super().__init__()

        if n_threads is None:
            n_threads = os.cpu_count()
            logger.warning(
                f"Engine set up to use up to {n_threads} processes in total. "
                f"The number was automatically determined and might not be "
                f"appropriate on some systems."
            )
        self.n_threads: int = n_threads

    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Deepcopy tasks and distribute work over parallel threads.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        n_tasks = len(tasks)

        copied_tasks = [copy.deepcopy(task) for task in tasks]

        n_threads = min(self.n_threads, n_tasks)
        logger.info(
            f"Performing parallel task execution on {n_threads} " f"threads."
        )

        with ThreadPoolExecutor(max_workers=n_threads) as pool:
            results = pool.map(
                work, tqdm(copied_tasks, disable=not progress_bar)
            )

        return results



"""
Ensemble
========
"""

from .covariance_analysis import (
    get_covariance_matrix_parameters,
    get_covariance_matrix_predictions,
    get_spectral_decomposition_lowlevel,
    get_spectral_decomposition_parameters,
    get_spectral_decomposition_predictions,
)
from .dimension_reduction import (
    get_pca_representation_parameters,
    get_pca_representation_predictions,
    get_umap_representation_parameters,
    get_umap_representation_predictions,
)
from .ensemble import Ensemble, EnsemblePrediction, get_percentile_label
from .util import (
    read_ensemble_prediction_from_h5,
    read_from_csv,
    read_from_df,
    write_ensemble_prediction_to_h5,
)



"""Abstract engine base class."""
import abc
from typing import List

from .task import Task


class Engine(abc.ABC):
    """Abstract engine base class."""

    def __init__(self):
        pass

    @abc.abstractmethod
    def execute(self, tasks: List[Task], progress_bar: bool = True):
        """Execute tasks.

        Parameters
        ----------
        tasks:
            List of tasks to execute.
        progress_bar:
            Whether to display a progress bar.
        """
        raise NotImplementedError("This engine is not intended to be called.")



"""Helper methods for hierarchical optimization with PEtab."""

from typing import Dict, Literal, Tuple

import pandas as pd
import petab
import sympy as sp
from more_itertools import one
from petab.C import LIN
from petab.C import LOWER_BOUND as PETAB_LOWER_BOUND
from petab.C import (
    OBSERVABLE_ID,
    OBSERVABLE_TRANSFORMATION,
    PARAMETER_SEPARATOR,
)
from petab.C import UPPER_BOUND as PETAB_UPPER_BOUND
from petab.observables import get_formula_placeholders

from ..C import INNER_PARAMETER_BOUNDS
from ..C import LOWER_BOUND as PYPESTO_LOWER_BOUND
from ..C import PARAMETER_TYPE
from ..C import UPPER_BOUND as PYPESTO_UPPER_BOUND
from ..C import InnerParameterType


def correct_parameter_df_bounds(parameter_df: pd.DataFrame) -> pd.DataFrame:
    """Correct the bounds of inner parameters in a PEtab parameters table.

    Parameters
    ----------
    parameter_df:
        The PEtab parameters table.

    Returns
    -------
    The table, with corrected bounds.
    """

    def correct_row(row: pd.Series) -> pd.Series:
        if pd.isna(row[PARAMETER_TYPE]):
            return row
        bounds = INNER_PARAMETER_BOUNDS[row[PARAMETER_TYPE]]
        row[PETAB_LOWER_BOUND] = bounds[PYPESTO_LOWER_BOUND]
        row[PETAB_UPPER_BOUND] = bounds[PYPESTO_UPPER_BOUND]
        return row

    return parameter_df.apply(correct_row, axis=1)


def validate_hierarchical_petab_problem(petab_problem: petab.Problem) -> None:
    """Validate a PEtab problem for hierarchical optimization.

    Parameters
    ----------
    petab_problem:
        The PEtab problem.
    """
    if PARAMETER_TYPE not in petab_problem.parameter_df:
        # not a hierarchical optimization problem
        return

    # ensure we only have linear parameter scale
    inner_parameter_table = petab_problem.parameter_df[
        petab_problem.parameter_df[PARAMETER_TYPE].isin(
            [
                InnerParameterType.OFFSET,
                InnerParameterType.SIGMA,
                InnerParameterType.SCALING,
            ]
        )
    ]
    if (
        petab.PARAMETER_SCALE in inner_parameter_table
        and not (
            inner_parameter_table[petab.PARAMETER_SCALE].isna()
            | (inner_parameter_table[petab.PARAMETER_SCALE] == petab.LIN)
        ).all()
    ):
        sub_df = inner_parameter_table.loc[
            :, [PARAMETER_TYPE, petab.PARAMETER_SCALE]
        ]
        raise NotImplementedError(
            "Only parameterScale=lin supported for parameters of the inner "
            f"subproblem.\n{sub_df}"
        )

    inner_parameter_df = validate_measurement_formulae(
        petab_problem=petab_problem
    )

    validate_inner_parameter_pairings(inner_parameter_df=inner_parameter_df)


def validate_inner_parameter_pairings(
    inner_parameter_df: pd.DataFrame,
):
    """Validate the pairings of inner parameters.

    Parameters
    ----------
    inner_parameter_df:
        Each row has the offset, scaling and sigma of each measurement in
        measurement table of the PEtab problem.
    """
    scalings_with_sigmas = {}
    offsets_with_sigmas = {}

    scalings_with_offsets = {}
    offsets_with_scalings = {}

    for _, row in inner_parameter_df.iterrows():
        offset = row[InnerParameterType.OFFSET]
        scaling = row[InnerParameterType.SCALING]
        sigma = row[InnerParameterType.SIGMA]

        # Ensures each scaling is only ever paired with one sigma.
        if scaling is not None and sigma is not None:
            expected_sigma = scalings_with_sigmas.get(scaling)
            if expected_sigma is not None and sigma != expected_sigma:
                raise ValueError(
                    "If a scaling parameter is paired with a sigma "
                    "parameter for one measurement, all measurements with "
                    "the same scaling parameter must not have a different "
                    "sigma inner parameter. "
                    f"Scaling: `{scaling}`. Sigma: `{sigma}`. "
                    f"Expected sigma: `{expected_sigma}`."
                )
            else:
                scalings_with_sigmas[scaling] = sigma

        # Ensures each offset is only ever paired with one sigma.
        if offset is not None and sigma is not None:
            expected_sigma = offsets_with_sigmas.get(offset)
            if expected_sigma is not None and sigma != expected_sigma:
                raise ValueError(
                    "If an offset parameter is paired with a sigma "
                    "parameter for one measurement, all measurements with "
                    "the same offset parameter must not have a different "
                    "sigma inner parameter. "
                    f"Scaling: `{offset}`. Sigma: `{sigma}`. "
                    f"Expected sigma: `{expected_sigma}`."
                )
            else:
                offsets_with_sigmas[offset] = sigma

        # Ensures each scaling is only ever paired with one offset, and vice
        # versa.
        if scaling is not None and offset is not None:
            expected_offset = scalings_with_offsets.get(scaling)
            expected_scaling = offsets_with_scalings.get(offset)
            if (expected_offset is not None and offset != expected_offset) or (
                expected_scaling is not None and scaling != expected_scaling
            ):
                raise ValueError(
                    "If a scaling parameter is paired with an offset "
                    "parameter for one measurement, all measurements with "
                    "the same scaling or offset parameter must not have a "
                    f"different scaling or offset inner parameter. "
                    f"Scaling: `{scaling}`. "
                    f"Expected scaling: `{expected_scaling}`. "
                    f"Offset: `{offset}`. "
                    f"Expected offset: `{expected_offset}`."
                )
            else:
                scalings_with_offsets[scaling] = offset
                offsets_with_scalings[offset] = scaling

        # TODO can a scaling that is paired with an offset, exist without
        #      any scaling?
        #      Same for offset?
        #      Same for scaling+sigma and offset+sigma pairings?


def get_inner_parameters(
    petab_problem: petab.Problem,
) -> Dict[str, InnerParameterType]:
    """Get information about the inner parameters.

    Parameters
    ----------
    petab_problem:
        The PEtab problem, with inner parameters.

    Returns
    -------
    Parameter IDs and their inner parameter types.
    """
    # TODO exclude fixed inner parameters

    if PARAMETER_TYPE not in petab_problem.parameter_df.columns:
        raise ValueError(
            "Hierarchical optimization requires specification of inner "
            f"parameters with the `{PARAMETER_TYPE}` column in the parameters "
            "table."
        )

    inner_parameters = {}
    for parameter_id, row in petab_problem.parameter_df.iterrows():
        type_str = row[PARAMETER_TYPE]
        if pd.isna(type_str):
            continue

        try:
            inner_parameters[parameter_id] = InnerParameterType(type_str)
        except ValueError as e:
            raise ValueError(
                f"Unknown inner parameter type `{type_str}`."
            ) from e

    return inner_parameters


def validate_measurement_formulae(
    petab_problem: petab.Problem,
) -> pd.DataFrame:
    """Check whether formulae associated with a measurement are valid.

    Specifically, check whether the measurement-specific observable and noise
    formulae are valid.

    Parameters
    ----------
    petab_problem:
        The PEtab problem.

    Returns
    -------
    A dataframe containing the inner parameters for each measurement.
    """
    inner_parameters = get_inner_parameters(petab_problem=petab_problem)

    inner_parameter_sets = []

    for _, measurement in petab_problem.measurement_df.iterrows():
        offset, scaling = _validate_measurement_specific_observable_formula(
            measurement=measurement,
            petab_problem=petab_problem,
            inner_parameters=inner_parameters,
        )
        sigma = _validate_measurement_specific_noise_formula(
            measurement=measurement,
            petab_problem=petab_problem,
            inner_parameters=inner_parameters,
        )
        inner_parameter_sets.append(
            [
                str(v) if v is not None else None
                for v in [offset, scaling, sigma]
            ]
        )

    return pd.DataFrame(
        data=inner_parameter_sets,
        columns=[
            InnerParameterType.OFFSET,
            InnerParameterType.SCALING,
            InnerParameterType.SIGMA,
        ],
    )


def _validate_measurement_specific_observable_formula(
    measurement: pd.Series,
    petab_problem: petab.Problem,
    inner_parameters: Dict[str, InnerParameterType],
) -> Tuple[InnerParameterType, InnerParameterType]:
    """Check whether a measurement observable formula is valid.

    Parameters
    ----------
    measurement:
        A row from a PEtab measurements table.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The offset and scaling parameters.
    """
    formula, formula_inner_parameters = _get_symbolic_formula_from_measurement(
        measurement=measurement,
        formula_type='observable',
        petab_problem=petab_problem,
        inner_parameters=inner_parameters,
    )

    offset = None
    scaling = None

    terms = sp.Add.make_args(formula)

    for (
        formula_inner_parameter,
        inner_parameter_type,
    ) in formula_inner_parameters.items():
        if inner_parameter_type == InnerParameterType.OFFSET:
            offset = formula_inner_parameter
            appearances = [
                term for term in terms if offset in term.free_symbols
            ]
            try:
                # `one` ensures the offset occurs in exactly one term
                # `==`  ensures that the offset appearances as its own term,
                #       e.g. without being multiplied by anything
                if offset != one(appearances):
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "An offset is in the observable formula, but is not of "
                    "the expected form. The observable formula should take "
                    "the form `y = ... + offset`. Observable formula: "
                    f"`{formula}`. Offset: `{offset}`."
                ) from e
        elif inner_parameter_type == InnerParameterType.SCALING:
            scaling = formula_inner_parameter
            appearances = [
                term for term in terms if scaling in term.free_symbols
            ]
            # Should occur in exactly one term, and should be multiplied by
            # everything else in the term.
            try:
                # `one` ensures the scaling occurs in exactly one term
                factors = sp.Mul.make_args(one(appearances))
                factor_appearances = [
                    factor
                    for factor in factors
                    if scaling in factor.free_symbols
                ]
                # as for offset, ensure scaling is exactly one factor
                if scaling != one(factor_appearances):
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "A scaling is in the observable formula, but is not of "
                    "the expected form. The observable formula should take "
                    "the form `y = scaling*(...) [+ offset]`. Observable "
                    f"formula: `{formula}`. Offset: `{scaling}`."
                ) from e
        else:
            raise ValueError("Unknown error: unexpected inner parameter type.")

    return offset, scaling


def _validate_measurement_specific_noise_formula(
    measurement: pd.Series,
    petab_problem: petab.Problem,
    inner_parameters: Dict[str, InnerParameterType],
) -> Tuple[InnerParameterType, InnerParameterType]:
    """Check whether a measurement noise formula is valid.

    Parameters
    ----------
    measurement:
        A row from a PEtab measurements table.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The sigma parameter.
    """
    formula, formula_inner_parameters = _get_symbolic_formula_from_measurement(
        measurement=measurement,
        formula_type='noise',
        petab_problem=petab_problem,
        inner_parameters=inner_parameters,
    )

    sigma = None

    for (
        formula_inner_parameter,
        inner_parameter_type,
    ) in formula_inner_parameters.items():
        if inner_parameter_type == InnerParameterType.SIGMA:
            sigma = formula_inner_parameter
            try:
                # The sigma must describe the noise completely, i.e. constitute
                # the full noise formula.
                if sigma != formula:
                    raise ValueError
            except Exception as e:
                raise ValueError(
                    "A sigma parameter does not constitute the full noise "
                    f"formula. Noise formula: `{formula}`. Sigma: `{sigma}`."
                ) from e
        else:
            raise ValueError("Unknown error: unexpected inner parameter type.")

    return sigma


def _get_symbolic_formula_from_measurement(
    measurement: pd.Series,
    formula_type: Literal['observable', 'noise'],
    petab_problem: petab.Problem,
    inner_parameters: Dict[str, InnerParameterType],
) -> Tuple[sp.Expr, Dict[sp.Symbol, InnerParameterType]]:
    """Get a symbolic representation of a formula, with overrides overridden.

    Also performs some checks to ensure only valid numbers and types of inner
    parameters are in the formula.

    Parameters
    ----------
    measurement:
        A row from the PEtab measurements table.
    formula_type:
        The formula to extract from the measurement.
    petab_problem:
        The PEtab problem.
    inner_parameters:
        See `get_inner_parameters`.

    Returns
    -------
    The symbolic formula, and its inner parameters.
    """
    observable_id = measurement[OBSERVABLE_ID]

    formula_string = petab_problem.observable_df.loc[
        observable_id, formula_type + 'Formula'
    ]
    symbolic_formula = sp.sympify(formula_string)

    formula_placeholders = get_formula_placeholders(
        formula_string=formula_string,
        observable_id=observable_id,
        override_type=formula_type,
    )
    if formula_placeholders:
        overrides = measurement[formula_type + 'Parameters']
        overrides = (
            overrides.split(PARAMETER_SEPARATOR)
            if isinstance(overrides, str)
            else [overrides]
        )
        subs = dict(zip(formula_placeholders, overrides))
        symbolic_formula = symbolic_formula.subs(subs)

    symbolic_formula_inner_parameters = {
        sp.Symbol(inner_parameter_id): inner_parameter_type
        for inner_parameter_id, inner_parameter_type in inner_parameters.items()
        if sp.Symbol(inner_parameter_id) in symbolic_formula.free_symbols
    }

    if formula_type == 'noise':
        max_parameters = 1
        expected_inner_parameter_types = [InnerParameterType.SIGMA]
    elif formula_type == 'observable':
        max_parameters = 2
        expected_inner_parameter_types = [
            InnerParameterType.OFFSET,
            InnerParameterType.SCALING,
        ]

    if len(symbolic_formula_inner_parameters) > max_parameters:
        raise ValueError(
            "There are too many inner parameters in the formula. "
            f"Formula type: `{formula_type}`. Formula: `{symbolic_formula}`. "
            f"Inner parameters: `{symbolic_formula_inner_parameters}`."
        )

    inner_parameter_types = set(symbolic_formula_inner_parameters.values())
    for inner_parameter_type in inner_parameter_types:
        if inner_parameter_type not in expected_inner_parameter_types:
            raise ValueError(
                f"Unexpected inner parameter type in {formula_type} formula. "
                f"Inner parameter type: `{inner_parameter_type}`. "
                f"Formula: `{symbolic_formula}`. Expected inner parameter "
                f"types: `{expected_inner_parameter_types}`."
            )
    if len(inner_parameter_types) != len(symbolic_formula_inner_parameters):
        raise ValueError(
            "There are multiple inner parameters of the same type."
            f"Inner parameters: `{symbolic_formula_inner_parameters.values}`."
        )

    if symbolic_formula_inner_parameters:
        observable_transformation = petab_problem.observable_df.loc[
            observable_id
        ].get(OBSERVABLE_TRANSFORMATION)
        if (
            observable_transformation is not None
            and observable_transformation != LIN
        ):
            raise ValueError(
                "Non-linear observable transformations are not supported if "
                "the observable is associated with hierarchically-optimized "
                f"inner parameters. "
                f"Observable transformation: `{observable_transformation}`. "
                f"Measurement:\n{measurement}"
            )

    return symbolic_formula, symbolic_formula_inner_parameters



from typing import Callable, Sequence, Tuple, Union

import numpy as np

from ..C import FVAL, GRAD, HESS, MODE_FUN, MODE_RES, RES, SRES, ModeType
from .base import ObjectiveBase, ResultDict


class Objective(ObjectiveBase):
    """
    Objective class.

    The objective class allows the user explicitly specify functions that
    compute the function value and/or residuals as well as respective
    derivatives.

    Denote dimensions `n` = parameters, `m` = residuals.

    Parameters
    ----------
    fun:
        The objective function to be minimized. If it only computes the
        objective function value, it should be of the form

            ``fun(x) -> float``

        where x is an 1-D array with shape (n,), and n is the parameter space
        dimension.
    grad:
        Method for computing the gradient vector. If it is a callable,
        it should be of the form

            ``grad(x) -> array_like, shape (n,).``

        If its value is True, then fun should return the gradient as a second
        output.
    hess:
        Method for computing the Hessian matrix. If it is a callable,
        it should be of the form

            ``hess(x) -> array, shape (n, n).``

        If its value is True, then fun should return the gradient as a
        second, and the Hessian as a third output, and grad should be True as
        well.
    hessp:
        Method for computing the Hessian vector product, i.e.

            ``hessp(x, v) -> array_like, shape (n,)``

        computes the product H*v of the Hessian of fun at x with v.
    res:
        Method for computing residuals, i.e.

            ``res(x) -> array_like, shape(m,).``
    sres:
        Method for computing residual sensitivities. If it is a callable,
        it should be of the form

            ``sres(x) -> array, shape (m, n).``

        If its value is True, then res should return the residual
        sensitivities as a second output.
    x_names:
        Parameter names. None if no names provided, otherwise a list of str,
        length dim_full (as in the Problem class). Can be read by the
        problem.
    """

    def __init__(
        self,
        fun: Callable = None,
        grad: Union[Callable, bool] = None,
        hess: Callable = None,
        hessp: Callable = None,
        res: Callable = None,
        sres: Union[Callable, bool] = None,
        x_names: Sequence[str] = None,
    ):
        self.fun = fun
        self.grad = grad
        self.hess = hess
        self.hessp = hessp
        self.res = res
        self.sres = sres
        super().__init__(x_names)

    @property
    def has_fun(self) -> bool:
        """Check whether function is defined."""
        return callable(self.fun)

    @property
    def has_grad(self) -> bool:
        """Check whether gradient is defined."""
        return callable(self.grad) or self.grad is True

    @property
    def has_hess(self) -> bool:
        """Check whether Hessian is defined."""
        return callable(self.hess) or self.hess is True

    @property
    def has_hessp(self) -> bool:
        """Check whether Hessian vector product is defined."""
        # Not supported yet
        return False

    @property
    def has_res(self) -> bool:
        """Check whether residuals are defined."""
        return callable(self.res)

    @property
    def has_sres(self) -> bool:
        """Check whether residual sensitivities are defined."""
        return callable(self.sres) or self.sres is True

    def get_config(self) -> dict:
        """Return basic information of the objective configuration."""
        info = super().get_config()
        info['x_names'] = self.x_names
        sensi_order = 0
        while self.check_sensi_orders(
            sensi_orders=(sensi_order,), mode=MODE_FUN
        ):
            sensi_order += 1
        info['sensi_order'] = sensi_order - 1
        return info

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        **kwargs,
    ) -> ResultDict:
        """
        Call objective function without pre- or post-processing and formatting.

        Returns
        -------
        result:
            A dict containing the results.
        """
        if mode == MODE_FUN:
            result = self._call_mode_fun(x=x, sensi_orders=sensi_orders)
        elif mode == MODE_RES:
            result = self._call_mode_res(x=x, sensi_orders=sensi_orders)
        else:
            raise ValueError("This mode is not supported.")
        return result

    def _call_mode_fun(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
    ) -> ResultDict:
        if not sensi_orders:
            result = {}

        elif sensi_orders == (0,):
            if self.grad is True:
                fval = self.fun(x)[0]
            else:
                fval = self.fun(x)
            result = {FVAL: fval}

        elif sensi_orders == (1,):
            if self.grad is True:
                grad = self.fun(x)[1]
            else:
                grad = self.grad(x)
            result = {GRAD: grad}
        elif sensi_orders == (2,):
            if self.hess is True:
                hess = self.fun(x)[2]
            else:
                hess = self.hess(x)
            result = {HESS: hess}
        elif sensi_orders == (0, 1):
            if self.grad is True:
                fval, grad = self.fun(x)[:2]
            else:
                fval = self.fun(x)
                grad = self.grad(x)
            result = {FVAL: fval, GRAD: grad}
        elif sensi_orders == (0, 2):
            if self.hess is True:
                fval, _, hess = self.fun(x)[:3]
            else:
                if self.grad is True:
                    fval = self.fun(x)[0]
                else:
                    fval = self.fun(x)
                hess = self.hess(x)
            result = {FVAL: fval, HESS: hess}
        elif sensi_orders == (1, 2):
            if self.hess is True:
                grad, hess = self.fun(x)[1:3]
            else:
                hess = self.hess(x)
                if self.grad is True:
                    grad = self.fun(x)[1]
                else:
                    grad = self.grad(x)
            result = {GRAD: grad, HESS: hess}
        elif sensi_orders == (0, 1, 2):
            if self.hess is True:
                fval, grad, hess = self.fun(x)[0:3]
            else:
                hess = self.hess(x)
                if self.grad is True:
                    fval, grad = self.fun(x)[0:2]
                else:
                    fval = self.fun(x)
                    grad = self.grad(x)
            result = {FVAL: fval, GRAD: grad, HESS: hess}
        else:
            raise ValueError("These sensitivity orders are not supported.")

        return result

    def _call_mode_res(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
    ) -> ResultDict:
        if not sensi_orders:
            result = {}

        elif sensi_orders == (0,):
            if self.sres is True:
                res = self.res(x)[0]
            else:
                res = self.res(x)
            result = {RES: res}

        elif sensi_orders == (1,):
            if self.sres is True:
                sres = self.res(x)[1]
            else:
                sres = self.sres(x)
            result = {SRES: sres}
        elif sensi_orders == (0, 1):
            if self.sres is True:
                res, sres = self.res(x)
            else:
                res = self.res(x)
                sres = self.sres(x)
            result = {RES: res, SRES: sres}
        else:
            raise ValueError("These sensitivity orders are not supported.")

        return result



"""Finite differences."""

import copy
import logging
from typing import Callable, Dict, List, Tuple, Union

import numpy as np

from ..C import FVAL, GRAD, HESS, MODE_FUN, MODE_RES, RES, SRES, ModeType
from .base import ObjectiveBase, ResultDict

logger = logging.getLogger(__name__)


class FDDelta:
    """Finite difference step size with automatic updating.

    Reference implementation:
    https://github.com/ICB-DCM/PESTO/blob/master/private/getStepSizeFD.m

    Parameters
    ----------
    delta:
        (Initial) step size, either a float, or a vector of size (n_par,).
        If not None, this is used as initial step size.
    test_deltas:
        Step sizes to try out in step size selection. If None, a range
        [1e-1, 1e-2, ..., 1e-8] is considered.
    update_condition:
        A "good" step size may be a local property. Thus, this class allows
        updating the step size if certain criteria are met, in the
        :func:`pypesto.objective.finite_difference.FDDelta.update`
        function.
        FDDelta.CONSTANT means that the step size is only initially selected.
        FDDelta.DISTANCE means that the step size is updated if the current
        evaluation point is sufficiently far away from the last training point.
        FDDelta.STEPS means that the step size is updated `max_steps`
        evaluations after the last update.
        FDDelta.ALWAYS mean that the step size is selected in every call.
    max_distance:
        Coefficient on the distance between current and reference point beyond
        which to update, in the `FDDelta.DISTANCE` update condition.
    max_steps:
        Number of steps after which to update in the `FDDelta.STEPS` update
        condition.
    """

    # update conditions
    CONSTANT = "constant"
    DISTANCE = "distance"
    STEPS = "steps"
    ALWAYS = "always"
    UPDATE_CONDITIONS = [CONSTANT, DISTANCE, STEPS, ALWAYS]

    def __init__(
        self,
        delta: Union[np.ndarray, float, None] = None,
        test_deltas: np.ndarray = None,
        update_condition: str = CONSTANT,
        max_distance: float = 0.5,
        max_steps: int = 30,
    ):
        if not isinstance(delta, (np.ndarray, float)) and delta is not None:
            raise ValueError(f"Unexpected type {type(delta)} for delta")
        self.delta: Union[np.ndarray, float, None] = delta

        if test_deltas is None:
            test_deltas = np.array([10 ** (-i) for i in range(1, 9)])
        self.test_deltas: np.ndarray = test_deltas

        if update_condition not in FDDelta.UPDATE_CONDITIONS:
            raise ValueError(
                f"Update condition {update_condition} must be in "
                f"{FDDelta.UPDATE_CONDITIONS}.",
            )
        self.update_condition: str = update_condition

        self.max_distance: float = max_distance
        self.max_steps: int = max_steps

        # run variables
        #  parameter where the step sizes where updated last
        self.x0: Union[np.ndarray, None] = None
        #  overall number of steps
        self.steps: int = 0
        self.updates: int = 0

    def update(
        self,
        x: np.ndarray,
        fval: Union[float, np.ndarray, None],
        fun: Callable,
        fd_method: str,
    ) -> None:
        """Update delta if update conditions are met.

        Parameters
        ----------
        x:
            Current parameter vector, shape (n_par,).
        fval:
            fun(x), to avoid re-evaluation. Scalar- or vector-valued.
        fun:
            Function whose 1st-order derivative to approximate.
            Scalar- or vector-valued.
        fd_method:
            FD method employed by
            :class:`pypesto.objective.finite_difference.FD`, see there.
        """
        # scalar to vector
        if isinstance(self.delta, float):
            self.delta: np.ndarray = self.delta * np.ones(shape=x.shape)
        elif isinstance(self.delta, np.ndarray):
            if self.delta.shape != x.shape:
                # this should not happen
                raise ValueError("Shape mismatch.")

        self.steps += 1

        # return if no update needed
        if self.delta is not None:
            if (
                (
                    self.update_condition == FDDelta.DISTANCE
                    and np.sum((x - self.x0) ** 2)
                    <= self.max_distance * np.sqrt(len(x))
                )
                or (
                    self.update_condition == FDDelta.STEPS
                    and (self.steps - 1) % self.max_steps != 0
                    and self.steps > 1
                )
                or (
                    self.update_condition == FDDelta.CONSTANT
                    and self.delta is not None
                )
            ):
                return

        # update reference point
        self.x0 = x
        # actually update
        self._update(x=x, fval=fval, fun=fun, fd_method=fd_method)

        if self.delta.shape != x.shape:
            # this should not happen
            raise ValueError("Shape mismatch.")

    def _update(
        self,
        x: np.ndarray,
        fval: Union[float, np.ndarray],
        fun: Callable,
        fd_method: str,
    ) -> None:
        """
        Actually update. Wants to be called in `update` explicitly.

        Run FDs with various deltas and pick the ones, separately for each
        parameter, with the best stability properties.

        The parameters are the same as for
        :func:`pypesto.objective.finite_difference.FDDelta.update`.
        """
        # calculate gradients for all deltas for all parameters
        nablas = []
        # iterate over deltas
        for delta in self.test_deltas:
            # calculate Jacobian with step size delta
            delta_vec = delta * np.ones_like(x)
            nabla = fd_nabla_1(
                x=x,
                fval=fval,
                f_fval=fun,
                delta_vec=delta_vec,
                fd_method=fd_method,
            )

            nablas.append(nabla)

        # shape (n_delta, n_par, ...)
        nablas = np.array(nablas)

        # The stability vector is the absolute difference of Jacobian
        #  entries towards smaller and larger deltas, thus indicating the
        #  change in the approximation when changing delta.
        # This is done separately for each parameter. Then, for each the delta
        #  with the minimal entry and thus the most stable behavior
        #  is selected.
        stab_vec = np.full(shape=nablas.shape, fill_value=np.nan)
        stab_vec[1:-1] = np.mean(
            np.abs([nablas[2:] - nablas[1:-1], nablas[1:-1] - nablas[:-2]]),
            axis=0,
        )
        # on the edge, just take the single neighbor
        stab_vec[0] = np.abs(nablas[1] - nablas[0])
        stab_vec[-1] = np.abs(nablas[-1] - nablas[-2])

        # if the function is tensor-valued, consider the maximum over all
        #  entries, to constrain the worst deviation
        if stab_vec.ndim > 2:
            # flatten all dimensions > 1
            stab_vec = stab_vec.reshape(
                stab_vec.shape[0], stab_vec.shape[1], -1
            ).max(axis=2)

        # minimum delta index for each parameter
        min_ixs = np.argmin(stab_vec, axis=0)

        # extract optimal deltas per parameter
        delta_opt = np.array([self.test_deltas[ix] for ix in min_ixs])
        self.delta = delta_opt

        # log
        logger.info(f"Optimal FD delta: {self.delta}")
        self.updates += 1

    def get(self) -> np.ndarray:
        """Get delta vector."""
        return self.delta


def to_delta(delta: Union[FDDelta, np.ndarray, float, str]) -> FDDelta:
    """Input to step size delta.

    Input can be a vector, float, or an update method type.

    Parameters
    ----------
    delta:
        Can be a vector, float, or one of Delta.UPDATE_CONDITIONS. If a
        vector or float, a constant delta is assumed.
    """
    if isinstance(delta, FDDelta):
        return delta
    elif isinstance(delta, (np.ndarray, float)):
        return FDDelta(delta=delta, update_condition=FDDelta.CONSTANT)
    else:
        return FDDelta(delta=None, update_condition=delta)


class FD(ObjectiveBase):
    """Finite differences (FDs) for derivatives.

    Given an objective that gives function values and/or residuals, this
    class allows to flexibly obtain all derivatives calculated via FDs.

    For the parameters `grad`, `hess`, `sres`, a value of None means that the
    objective derivative is used if available, otherwise resorting to FDs.
    True means that FDs are used in any case, False means that the derivative
    is not exported.

    Note that the step sizes should be carefully chosen. They should be small
    enough to provide an accurate linear approximation, but large enough to
    be robust against numerical inaccuracies, in particular if the objective
    relies on numerical approximations, such as an ODE.

    Parameters
    ----------
    grad:
        Derivative method for the gradient (see above).
    hess:
        Derivative method for the Hessian (see above).
    sres:
        Derivative method for the residual sensitivities (see above).
    hess_via_fval:
        If the Hessian is to be calculated via finite differences:
        whether to employ 2nd order FDs via fval even if the objective can
        provide a gradient.
    delta_fun:
        FD step sizes for function values.
        Can be either a float, or a :class:`np.ndarray` of shape (n_par,)
        for different step sizes for different coordinates.
    delta_grad:
        FD step sizes for gradients, if the Hessian is calculated via 1st
        order sensitivities from the gradients. Similar to `delta_fun`.
    delta_res:
        FD step sizes for residuals. Similar to `delta_fun`.
    method:
        Method to calculate FDs. Can be any of `FD.METHODS`: central,
        forward or backward differences. The latter two require only roughly
        half as many function evaluations, are however less accurate than
        central (O(x) vs O(x**2)).
    x_names:
        Parameter names that can be optionally used in, e.g., history or
        gradient checks.

    Examples
    --------
    Define residuals and objective function, and obtain all derivatives via
    FDs:

    >>> from pypesto import Objective, FD
    >>> import numpy as np
    >>> x_obs = np.array([11, 12, 13])
    >>> res = lambda x: x - x_obs
    >>> fun = lambda x: 0.5 * sum(res(x)**2)
    >>> obj = FD(Objective(fun=fun, res=res))
    """

    # finite difference types
    CENTRAL = "central"
    FORWARD = "forward"
    BACKWARD = "backward"
    METHODS = [CENTRAL, FORWARD, BACKWARD]

    def __init__(
        self,
        obj: ObjectiveBase,
        grad: Union[bool, None] = None,
        hess: Union[bool, None] = None,
        sres: Union[bool, None] = None,
        hess_via_fval: bool = True,
        delta_fun: Union[FDDelta, np.ndarray, float, str] = 1e-6,
        delta_grad: Union[FDDelta, np.ndarray, float, str] = 1e-6,
        delta_res: Union[FDDelta, float, np.ndarray, str] = 1e-6,
        method: str = CENTRAL,
        x_names: List[str] = None,
    ):
        super().__init__(x_names=x_names)
        self.obj: ObjectiveBase = obj
        self.grad: Union[bool, None] = grad
        self.hess: Union[bool, None] = hess
        self.sres: Union[bool, None] = sres
        self.hess_via_fval: bool = hess_via_fval
        self.delta_fun: FDDelta = to_delta(delta_fun)
        self.delta_grad: FDDelta = to_delta(delta_grad)
        self.delta_res: FDDelta = to_delta(delta_res)
        self.method: str = method

        if method not in FD.METHODS:
            raise ValueError(
                f"Method must be one of {FD.METHODS}.",
            )

    def __deepcopy__(
        self,
        memodict: Dict = None,
    ) -> 'FD':
        """Create deepcopy of Objective."""
        other = self.__class__.__new__(self.__class__)
        for attr, val in self.__dict__.items():
            other.__dict__[attr] = copy.deepcopy(val)
        return other

    @property
    def has_fun(self) -> bool:
        """Check whether function is defined."""
        return self.obj.has_fun

    @property
    def has_grad(self) -> bool:
        """Check whether gradient is defined."""
        return self.grad is not False and self.obj.has_fun

    @property
    def has_hess(self) -> bool:
        """Check whether Hessian is defined."""
        return self.hess is not False and self.obj.has_fun

    @property
    def has_res(self) -> bool:
        """Check whether residuals are defined."""
        return self.obj.has_res

    @property
    def has_sres(self) -> bool:
        """Check whether residual sensitivities are defined."""
        return self.sres is not False and self.obj.has_res

    def call_unprocessed(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        mode: ModeType,
        **kwargs,
    ) -> ResultDict:
        """
        See `ObjectiveBase` for more documentation.

        Main method to overwrite from the base class. It handles and
        delegates the actual objective evaluation.
        """
        if mode == MODE_FUN:
            result = self._call_mode_fun(
                x=x, sensi_orders=sensi_orders, **kwargs
            )
        elif mode == MODE_RES:
            result = self._call_mode_res(
                x=x, sensi_orders=sensi_orders, **kwargs
            )
        else:
            raise ValueError("This mode is not supported.")

        return result

    def _call_mode_fun(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        **kwargs,
    ) -> ResultDict:
        """Handle calls in function value mode.

        Delegated from `call_unprocessed`.
        """
        # get from objective what it can and should deliver
        sensi_orders_obj, result = self._call_from_obj_fun(
            x=x,
            sensi_orders=sensi_orders,
            **kwargs,
        )

        # remaining sensis via FDs

        # whether gradient and Hessian are intended as FDs
        grad_via_fd = 1 in sensi_orders and 1 not in sensi_orders_obj
        hess_via_fd = 2 in sensi_orders and 2 not in sensi_orders_obj

        if not grad_via_fd and not hess_via_fd:
            return result

        # whether the Hessian should be based on 2nd order FD from fval
        hess_via_fd_fval = hess_via_fd and (
            self.hess_via_fval or not self.obj.has_grad
        )
        hess_via_fd_grad = hess_via_fd and not hess_via_fd_fval

        def f_fval(x):
            """Short-hand to get a function value."""
            return self.obj.call_unprocessed(
                x=x, sensi_orders=(0,), mode=MODE_FUN, **kwargs
            )[FVAL]

        def f_grad(x):
            """Short-hand to get a gradient value."""
            return self.obj.call_unprocessed(
                x=x, sensi_orders=(1,), mode=MODE_FUN, **kwargs
            )[GRAD]

        # update delta vectors
        if grad_via_fd or hess_via_fd_fval:
            # note: we use the same delta for 1st and 2nd order approximations
            # this may be not ideal
            self.delta_fun.update(
                x=x, fval=result.get(FVAL), fun=f_fval, fd_method=self.method
            )
        if hess_via_fd_grad:
            self.delta_grad.update(
                x=x, fval=result.get(GRAD), fun=f_grad, fd_method=self.method
            )

        # calculate gradient
        if grad_via_fd:
            result[GRAD] = fd_nabla_1(
                x=x,
                fval=result.get(FVAL),
                f_fval=f_fval,
                delta_vec=self.delta_fun.get(),
                fd_method=self.method,
            )

        # calculate Hessian
        if hess_via_fd:
            if hess_via_fd_fval:
                result[HESS] = fd_nabla_2(
                    x=x,
                    fval=result.get(FVAL),
                    f_fval=f_fval,
                    delta_vec=self.delta_fun.get(),
                    fd_method=self.method,
                )
            else:
                hess = fd_nabla_1(
                    x=x,
                    fval=result.get(GRAD),
                    f_fval=f_grad,
                    delta_vec=self.delta_fun.get(),
                    fd_method=self.method,
                )
                # make it symmetric
                result[HESS] = 0.5 * (hess + hess.T)

        return result

    def _call_mode_res(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        **kwargs,
    ) -> ResultDict:
        """Handle calls in residual mode.

        Delegated from `call_unprocessed`.
        """
        # get from objective what it can and should deliver
        sensi_orders_obj, result = self._call_from_obj_res(
            x=x,
            sensi_orders=sensi_orders,
            **kwargs,
        )

        if sensi_orders == sensi_orders_obj:
            # all done
            return result

        def f_res(x):
            """Short-hand to get a function value."""
            return self.obj.call_unprocessed(
                x=x, sensi_orders=(0,), mode=MODE_RES, **kwargs
            )[RES]

        # update delta vector
        self.delta_res.update(
            x=x, fval=result.get(RES), fun=f_res, fd_method=self.method
        )

        # sres
        sres = fd_nabla_1(
            x=x,
            fval=result.get(RES),
            f_fval=f_res,
            delta_vec=self.delta_res.get(),
            fd_method=self.method,
        )
        # sres should have shape (n_res, n_par)
        result[SRES] = sres.T

        return result

    def _call_from_obj_fun(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        **kwargs,
    ) -> Tuple[Tuple[int, ...], ResultDict]:
        """
        Call objective function for sensitivities.

        Calculate from the objective the sensitivities that are supposed to
        be calculated from the objective and not via FDs.
        """
        # define objective sensis
        sensi_orders_obj = []
        if 0 in sensi_orders:
            sensi_orders_obj.append(0)
        if 1 in sensi_orders and self.grad is None and self.obj.has_grad:
            sensi_orders_obj.append(1)
        if 2 in sensi_orders and self.hess is None and self.obj.has_hess:
            sensi_orders_obj.append(2)
        sensi_orders_obj = tuple(sensi_orders_obj)
        # call objective
        result = {}
        if sensi_orders_obj:
            result = self.obj.call_unprocessed(
                x=x, sensi_orders=sensi_orders_obj, mode=MODE_FUN, **kwargs
            )
        return sensi_orders_obj, result

    def _call_from_obj_res(
        self,
        x: np.ndarray,
        sensi_orders: Tuple[int, ...],
        **kwargs,
    ) -> Tuple[Tuple[int, ...], ResultDict]:
        """
        Call objective function for sensitivities in residual mode.

        Calculate from the objective the sensitivities that are supposed to
        be calculated from the objective and not via FDs.
        """
        # define objective sensis
        sensi_orders_obj = []
        if 0 in sensi_orders:
            sensi_orders_obj.append(0)
        if 1 in sensi_orders and self.sres is None and self.obj.has_sres:
            sensi_orders_obj.append(1)
        sensi_orders_obj = tuple(sensi_orders_obj)
        #  call objective
        result = {}
        if sensi_orders_obj:
            result = self.obj.call_unprocessed(
                x=x, sensi_orders=sensi_orders_obj, mode=MODE_RES, **kwargs
            )
        return sensi_orders_obj, result


def unit_vec(dim: int, ix: int) -> np.ndarray:
    """
    Return unit vector of dimension `dim` at coordinate `ix`.

    Parameters
    ----------
    dim: Vector dimension.
    ix: Index to contain the unit value.

    Returns
    -------
    vector: The unit vector.
    """
    vector = np.zeros(shape=dim)
    vector[ix] = 1
    return vector


def fd_nabla_1(
    x: np.ndarray,
    fval: float,
    f_fval: Callable,
    delta_vec: np.ndarray,
    fd_method: str,
) -> np.ndarray:
    """Calculate FD approximation to 1st order derivative (Jacobian/Gradient).

    Parameters
    ----------
    x: Parameter vector, shape (n_par,).
    fval: Function value at `x`, calculated if None.
    f_fval: Function returning function values. Scalar- or vector-valued.
    delta_vec: Step size vector, shape (n_par,).
    fd_method: FD method.

    Keyword arguments are passed on to the objective.

    Returns
    -------
    nabla_1:
        The FD approximation to the 1st order derivatives.
        Shape (n_par, ...) with ndim > 1 if `f_fval` is not scalar-valued.
    """
    # parameter dimension
    n_par = len(x)

    # calculate value at x only once if needed
    if fval is None and fd_method in [FD.FORWARD, FD.BACKWARD]:
        fval = f_fval(x)

    nabla_1 = []
    for ix in range(n_par):
        delta_val = delta_vec[ix]
        delta = delta_val * unit_vec(dim=n_par, ix=ix)

        if fd_method == FD.CENTRAL:
            fp = f_fval(x + delta / 2)
            fm = f_fval(x - delta / 2)
        elif fd_method == FD.FORWARD:
            fp = f_fval(x + delta)
            fm = fval
        elif fd_method == FD.BACKWARD:
            fp = fval
            fm = f_fval(x - delta)
        else:
            raise ValueError("Method not recognized.")

        nabla_1.append((fp - fm) / delta_val)

    return np.array(nabla_1)


def fd_nabla_2(
    x: np.ndarray,
    fval: float,
    f_fval: Callable,
    delta_vec: np.ndarray,
    fd_method: str,
) -> np.ndarray:
    """Calculate FD approximation to 2nd order derivatives (e.g. Hessian).

    Parameters
    ----------
    x: Parameter vector, shape (n_par,).
    fval: Function value at `x`, calculated if None. Scalar- or vector-valued.
    f_fval: Function returning function values.
    delta_vec: Step size vector, shape (n_par,).
    fd_method: FD method.

    Returns
    -------
    nabla_2:
        The FD approximation of the 2nd order derivative tensor.
        Shape (n_par, n_par, ...) with ndim > 2 if `f_fval` is not
        scalar-valued.
    """
    # parameter dimension
    n_par = len(x)

    # needed for diagonal entries at least
    if fval is None:
        fval = f_fval(x)

    # create empty matrix
    nabla_2 = []
    for _ in range(n_par):
        nabla_2.append([None] * n_par)

    # iterate over all parameter index tuples
    for ix1 in range(n_par):

        delta1_val = delta_vec[ix1]
        delta1 = delta1_val * unit_vec(dim=n_par, ix=ix1)

        # diagonal entry
        if fd_method == FD.CENTRAL:
            f2p = f_fval(x + delta1)
            fc = fval
            f2m = f_fval(x - delta1)
        elif fd_method == FD.FORWARD:
            f2p = f_fval(x + 2 * delta1)
            fc = f_fval(x + delta1)
            f2m = fval
        elif fd_method == FD.BACKWARD:
            f2p = fval
            fc = f_fval(x - delta1)
            f2m = f_fval(x - 2 * delta1)
        else:
            raise ValueError(f"Method {fd_method} not recognized.")

        nabla_2[ix1][ix1] = (f2p + f2m - 2 * fc) / delta1_val**2

        # off-diagonals
        for ix2 in range(ix1):
            delta2_val = delta_vec[ix2]
            delta2 = delta2_val * unit_vec(dim=n_par, ix=ix2)

            if fd_method == FD.CENTRAL:
                fpp = f_fval(x + delta1 / 2 + delta2 / 2)
                fpm = f_fval(x + delta1 / 2 - delta2 / 2)
                fmp = f_fval(x - delta1 / 2 + delta2 / 2)
                fmm = f_fval(x - delta1 / 2 - delta2 / 2)
            elif fd_method == FD.FORWARD:
                fpp = f_fval(x + delta1 + delta2)
                fpm = f_fval(x + delta1 + 0)
                fmp = f_fval(x + 0 + delta2)
                fmm = fval
            elif fd_method == FD.BACKWARD:
                fpp = fval
                fpm = f_fval(x + 0 - delta2)
                fmp = f_fval(x - delta1 + 0)
                fmm = f_fval(x - delta1 - delta2)
            else:
                raise ValueError(f"Method {fd_method} not recognized.")

            nabla_2[ix1][ix2] = nabla_2[ix2][ix1] = (fpp - fpm - fmp + fmm) / (
                delta1_val * delta2_val
            )

    return np.array(nabla_2)



from typing import Sequence, Tuple, Union
from warnings import warn

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator

from ..result import Result
from .clust_color import assign_colors
from .misc import process_result_list
from .reference_points import ReferencePoint, create_references


def profiles(
    results: Union[Result, Sequence[Result]],
    ax=None,
    profile_indices: Sequence[int] = None,
    size: Sequence[float] = (18.5, 6.5),
    reference: Union[ReferencePoint, Sequence[ReferencePoint]] = None,
    colors=None,
    legends: Sequence[str] = None,
    x_labels: Sequence[str] = None,
    profile_list_ids: Union[int, Sequence[int]] = 0,
    ratio_min: float = 0.0,
    show_bounds: bool = False,
):
    """
    Plot classical 1D profile plot.

    Using the posterior, e.g. Gaussian like profile.

    Parameters
    ----------
    results: list or pypesto.Result
        List of or single `pypesto.Result` after profiling.
    ax: list of matplotlib.Axes, optional
        List of axes objects to use.
    profile_indices: list of integer values
        List of integer values specifying which profiles should be plotted.
    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified.
    reference: list, optional
        List of reference points for optimization results, containing at
        least a function value fval.
    colors: list, or RGBA, optional
        List of colors, or single color.
    legends: list or str, optional
        Labels for line plots, one label per result object.
    x_labels: list of str
        Labels for parameter value axes (e.g. parameter names).
    profile_list_ids: int or list of ints, optional
        Index or list of indices of the profile lists to be used for profiling.
    ratio_min:
        Minimum ratio below which to cut off.
    show_bounds:
        Whether to show, and extend the plot to, the lower and upper bounds.

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # parse input
    results, profile_list_ids, colors, legends = process_result_list_profiles(
        results, profile_list_ids, colors, legends
    )

    # get the parameter ids to be plotted
    profile_indices = process_profile_indices(
        results, profile_indices, profile_list_ids
    )

    # loop over results
    for i_result, result in enumerate(results):
        for i_profile_list, profile_list_id in enumerate(profile_list_ids):
            fvals = handle_inputs(
                result,
                profile_indices=profile_indices,
                profile_list=profile_list_id,
                ratio_min=ratio_min,
            )

            # add x_labels for parameters
            if x_labels is None:
                x_labels = [
                    name
                    for name, fval in zip(result.problem.x_names, fvals)
                    if fval is not None
                ]

            # plot multiple results or profile runs into one figure?
            if len(results) == 1 and len(profile_list_ids) > 1:
                # multiple profile runs per axes object
                color_ind = i_profile_list
            else:
                # multiple results per axes object
                color_ind = i_result

            # call lowlevel routine
            ax = profiles_lowlevel(
                fvals=fvals,
                ax=ax,
                size=size,
                color=colors[color_ind],
                legend_text=legends[color_ind],
                x_labels=x_labels,
                show_bounds=show_bounds,
                lb_full=result.problem.lb_full,
                ub_full=result.problem.ub_full,
            )

    # parse and apply plotting options
    ref = create_references(references=reference)

    # plot reference points
    ax = handle_reference_points(ref, ax, profile_indices)

    plt.tight_layout()

    return ax


def profiles_lowlevel(
    fvals,
    ax=None,
    size: Tuple[float, float] = (18.5, 6.5),
    color=None,
    legend_text: str = None,
    x_labels=None,
    show_bounds: bool = False,
    lb_full=None,
    ub_full=None,
):
    """
    Lowlevel routine for profile plotting.

    Working with a list of arrays only, opening different axes objects in case.

    Parameters
    ----------
    fvals: numeric list or array
        Values to plot.
    ax: list of matplotlib.Axes, optional
        List of axes object to use.
    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified.
    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified.
    color: RGBA, optional
        Color for profiles in plot.
    legend_text: str
        Label for line plots.
    legend_text: List[str]
        Label for line plots.
    show_bounds:
        Whether to show, and extend the plot to, the lower and upper bounds.
    lb_full:
        Lower bound.
    ub_full:
        Upper bound.

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # axes
    if ax is None:
        ax = []
        fig = plt.figure()
        fig.set_size_inches(*size)
        create_new_ax = True
    else:
        plt.axes(ax[0])
        fig = plt.gcf()
        create_new_ax = False

    # count number of necessary axes
    if isinstance(fvals, list):
        n_fvals = len(fvals)
    else:
        n_fvals = 1
        fvals = [fvals]

    # number of non-trivial profiles
    n_profiles = sum((fval is not None for fval in fvals))

    # if axes already exists, we have to match profiles to axes
    if not create_new_ax:
        if n_fvals != len(ax) and n_profiles != len(ax):
            raise ValueError(
                "Number of axes does not match number of profiles. Stopping."
            )
        elif n_fvals == len(ax) and n_profiles != len(ax):
            # we may have some empty profiles, which we have to skip
            n_plots = n_fvals
        else:
            # n_profiles == len(ax):, we have exactly as many profiles as axes
            n_plots = n_profiles
    else:
        n_plots = n_profiles

    if lb_full is None:
        lb_full = [None] * len(fvals)
    if ub_full is None:
        ub_full = [None] * len(fvals)

    # compute number of columns and rows
    columns = np.ceil(np.sqrt(n_plots))
    if n_plots > columns * (columns - 1):
        rows = columns
    else:
        rows = columns - 1

    counter = 0
    for i_plot, (fval, lb, ub) in enumerate(zip(fvals, lb_full, ub_full)):
        # if we have empty profiles and more axes than profiles: skip
        if n_plots != n_fvals and fval is None:
            continue

        # handle legend
        if i_plot == 0:
            tmp_legend = legend_text
        else:
            tmp_legend = None

        # create or choose an axes object
        if create_new_ax:
            ax.append(fig.add_subplot(int(rows), int(columns), counter + 1))
        else:
            plt.axes(ax[counter])

        # plot if data
        if fval is not None:
            # run lowlevel routine for one profile
            ax[counter] = profile_lowlevel(
                fval,
                ax[counter],
                size=size,
                color=color,
                legend_text=tmp_legend,
                show_bounds=show_bounds,
                lb=lb,
                ub=ub,
            )

        # labels
        if x_labels is None:
            ax[counter].set_xlabel(f'Parameter {i_plot}')
        else:
            ax[counter].set_xlabel(x_labels[counter])

        if counter % columns == 0:
            ax[counter].set_ylabel('Log-posterior ratio')
        else:
            # fix pyPESTO/pyPESTO/pypesto/visualize/profiles.py:228:
            # UserWarning: FixedFormatter should only be used
            # together with FixedLocator. Fix from matplotlib #18848.
            ax[counter].set_yticks(ax[counter].get_yticks())
            ax[counter].set_yticklabels(['' for _ in ax[counter].get_yticks()])

        # increase counter and cleanup legend
        counter += 1

    return ax


def profile_lowlevel(
    fvals,
    ax=None,
    size: Tuple[float, float] = (18.5, 6.5),
    color=None,
    legend_text: str = None,
    show_bounds: bool = False,
    lb: float = None,
    ub: float = None,
):
    """
    Lowlevel routine for plotting one profile, working with a numpy array only.

    Parameters
    ----------
    fvals: numeric list or array
        Values to plot.
    ax: matplotlib.Axes, optional
        Axes object to use.
    size: tuple, optional
        Figure size (width, height) in inches. Is only applied when no ax
        object is specified.
    color: RGBA, optional
        Color for profiles in plot.
    legend_text: str
        Label for line plots.
    show_bounds:
        Whether to show, and extend the plot to, the lower and upper bounds.
    lb:
        Lower bound.
    ub:
        Upper bound.

    Returns
    -------
    ax: matplotlib.Axes
        The plot axes.
    """
    # parse input
    fvals = np.asarray(fvals)

    # get colors
    color = assign_colors([1.0], color)

    # axes
    if ax is None:
        ax = plt.subplots()[1]
        ax.set_xlabel('Parameter value')
        ax.set_ylabel('Log-posterior ratio')
        fig = plt.gcf()
        fig.set_size_inches(*size)

    # plot
    if fvals.size != 0:
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.plot(fvals[0, :], fvals[1, :], color=color[0], label=legend_text)

    if legend_text is not None:
        ax.legend()

    if show_bounds:
        ax.set_xlim([lb, ub])

    return ax


def handle_reference_points(ref, ax, profile_indices):
    """
    Handle reference points.

    Parameters
    ----------
    ref: list, optional
        List of reference points for optimization results, containing et
        least a function value fval
    ax: matplotlib.Axes, optional
        Axes object to use.
    profile_indices: list of integer values
        List of integer values specifying which profiles should be plotted.
    """
    if len(ref) > 0:
        # loop over axes objects
        for i_par, i_ax in enumerate(ax):
            for i_ref in ref:
                current_x = i_ref['x'][profile_indices[i_par]]
                i_ax.plot(
                    [current_x, current_x],
                    [0.0, 1.0],
                    color=i_ref.color,
                    label=i_ref.legend,
                )

            # create legend for reference points
            if i_ref.legend is not None:
                i_ax.legend()

    return ax


def handle_inputs(
    result: Result,
    profile_indices: Sequence[int],
    profile_list: int,
    ratio_min: float,
):
    """
    Retrieve the values of the profiles to be plotted.

    Parameters
    ----------
    result: pypesto.Result
        Profile result obtained by 'profile.py'.
    profile_indices: list of integer values
        List of integer values specifying which profiles should be plotted.
    profile_list: int, optional
        Index of the profile list to be used for profiling.
    ratio_min: int, optional
        Exclude values where profile likelihood ratio is smaller than
        ratio_min.

    Returns
    -------
    fvals: numeric list
        Including values that need to be plotted.
    """
    # extract ratio values values from result
    fvals = []
    for i_par in range(0, len(result.profile_result.list[profile_list])):
        if (
            i_par in profile_indices
            and result.profile_result.list[profile_list][i_par] is not None
        ):
            xs = result.profile_result.list[profile_list][i_par].x_path[
                i_par, :
            ]
            ratios = result.profile_result.list[profile_list][
                i_par
            ].ratio_path[:]

            # constrain
            indices = np.where(ratios > ratio_min)
            xs = xs[indices]
            ratios = ratios[indices]

            fvals_for_par = np.array([xs, ratios])
        else:
            fvals_for_par = None
        fvals.append(fvals_for_par)

    return fvals


def process_result_list_profiles(
    results: Result,
    profile_list_ids: Sequence[int],
    colors: Sequence[np.array],
    legends: Union[str, list],
) -> Sequence[int]:
    """
    Assign colors and legends to a list of results.

    Takes also care of the special cases for profile plotting.

    Parameters
    ----------
    results: list or pypesto.Result
        List of or single `pypesto.Result` after profiling.
    profile_list_ids: int or list of ints, optional
        Index or list of indices of the profile lists to be used for profiling.
    colors: list of RGBA colors
        colors for
    legends: list of str
        Legends for plotting

    Returns
    -------
    profile_indices: list of integer values
        corrected list of integer values specifying which profiles should be
        plotted.
    """
    # ensure list of ids
    if isinstance(profile_list_ids, int):
        profile_list_ids = [profile_list_ids]

    # check if we have a single result
    if isinstance(results, list):
        if len(results) != 1:
            # if we have no single result, then use the standard api
            results, colors, legends = process_result_list(
                results, colors, legends
            )
            return results, profile_list_ids, colors, legends
    else:
        # a single results was provided, so make a list out of it
        results = [results]

    # If we have a single result, we may still have multiple profile_list_ids
    # which should be plotted separately: use profile_list_ids as results dummy
    _, colors, legends = process_result_list(profile_list_ids, colors, legends)

    return results, profile_list_ids, colors, legends


def process_profile_indices(
    results: Sequence[Result],
    profile_indices: Sequence[int],
    profile_list_ids: Union[int, Sequence[int]],
):
    """
    Clean up profile_indices to be plotted.

    Retrieve the indices of the parameter for which profiles should be
    plotted later from a list of pypesto.ProfileResult objects.
    """
    # get all parameter indices, for which profiles were computed
    plottable_indices = set()
    for result in results:
        for profile_list_id in profile_list_ids:
            # get parameter indices, for which profiles were computed
            if profile_list_id < len(result.profile_result.list):
                tmp_indices = [
                    par_id
                    for par_id, prof in enumerate(
                        result.profile_result.list[profile_list_id]
                    )
                    if prof is not None
                ]
                # profile_indices should contain all parameter indices,
                # for which in at least one of the results a profile exists
                plottable_indices.update(tmp_indices)
    plottable_indices = sorted(plottable_indices)

    # get the profiles, which should be plotted and sanitize, if not plottable
    if profile_indices is None:
        profile_indices = list(plottable_indices)
    else:
        for ind in profile_indices:
            if ind not in plottable_indices:
                profile_indices.remove(ind)
                warn(
                    'Requested to plot profile for parameter index %i, '
                    'but profile has not been computed.' % ind
                )

    return profile_indices



"""Helper function for `geweke_test`."""
import logging
import warnings
from typing import Tuple

import numpy as np
from scipy.stats import norm

logger = logging.getLogger(__name__)


def spectrum(x: np.ndarray, nfft: int = None, nw: int = None) -> np.ndarray:
    """
    Power spectral density using Hanning window.

    Parameters
    ----------
    x:
        Fraction/fragment of one single parameter
        of the MCMC chain.
    nfft:
        The n-point discrete Fourier Transform.
    nw:
        Number of windows.

    Returns
    -------
    spectral_density:
        The spectral density.
    """
    if nfft is None:
        nfft = np.min(len(x), 256)

    if nw is None:
        nw = np.floor(nfft / 4).astype(int)

    n_overlap = np.floor(nw / 2).astype(int)

    # Hanning window
    w = 0.5 * (
        1 - np.cos(2 * np.pi * np.transpose(np.arange(1, nw + 1)) / (nw + 1))
    )
    n = len(x)
    if n < nw:
        x[nw] = 0
        n = nw

    # Number of windows
    k = np.floor((n - n_overlap) / (nw - n_overlap)).astype(int)
    index = np.arange(nw)
    # Normalizing scale factor
    kmu = k * np.linalg.norm(w) ** 2
    spectral_density = np.zeros((nfft))

    for _ in range(k):
        xw = w * x[index]
        index += nw - n_overlap
        Xx = np.absolute(np.fft.fft(xw, n=nfft, axis=0)) ** 2
        spectral_density += Xx

    # Normalize
    spectral_density = spectral_density * (1 / kmu)

    n2 = np.floor(nfft / 2).astype(int)

    spectral_density = spectral_density[0:n2]

    return spectral_density


def spectrum0(x: np.ndarray) -> np.ndarray:
    """
    Calculate the spectral density at frequency zero.

    Parameters
    ----------
    x:
        Fraction/fragment of the MCMC chain.

    Returns
    -------
    spectral_density_zero:
        Spectral density at zero.
    """
    n_samples, n_par = x.shape
    spectral_density_zero = np.zeros((1, n_par))

    for i in range(n_par):
        _spectral_density_zero = spectrum(x[:, i], n_samples)
        if len(_spectral_density_zero) > 0:
            spectral_density_zero[:, i] = _spectral_density_zero[0]
    return spectral_density_zero


def calculate_zscore(
    chain: np.ndarray, a: float = 0.1, b: float = 0.5
) -> Tuple[float, float]:
    """
    Perform a Geweke test on a chain.

    Use the first "a" fraction and the last "b" fraction of it for
    comparison. Test for equality of the means of the first a% and last b%
    of a Markov chain.

    See:
    Stephen P. Brooks and Gareth O. Roberts.
    Assessing convergence of Markov chain Monte Carlo
    algorithms. Statistics and Computing, 8:319--335, 1998.

    Parameters
    ----------
    chain
    a:
        First fraction of the MCMC chain.
    b:
        Second fraction of the MCMC chain.

    Returns
    -------
    z_score:
        Z-score of the Geweke test.
    p:
        Significance level of the Geweke test.
    """
    nsamples, _ = chain.shape

    # Define First fraction
    index_a = np.floor(a * nsamples).astype(int)
    # Define Second fraction
    index_b = nsamples - np.floor(b * nsamples).astype(int) + 1

    # Check if appropiate indexes
    if (index_a + index_b) / nsamples > 1:
        raise ValueError(
            "Sample size too small to "
            "meaningfully extract subsets "
            "for Geweke's test."
        )

    # Expect to see RuntimeWarnings in this block for short chains
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        # Mean of First fraction
        mean_a = np.mean(chain[0:index_a, :], axis=0)
        # Mean of Second fraction
        mean_b = np.mean(chain[index_b:, :], axis=0)

    # Spectral estimates for variance
    spectrum_a = spectrum0(chain[0:index_a, :])
    spectrum_b = spectrum0(chain[index_b:, :])

    # Calculate z-score
    z_score = (mean_a - mean_b) / (
        np.sqrt(spectrum_a / index_a + spectrum_b / (nsamples - index_b + 1))
    )
    # Calculate significance (p value)
    p = 2 * (1 - norm.cdf(np.absolute(z_score)))

    return z_score, p


def burn_in_by_sequential_geweke(
    chain: np.ndarray, zscore: float = 2.0
) -> int:
    """
    Calculate the burn-in of MCMC chains.

    Parameters
    ----------
    chain:
        The MCMC chain after removing warm up phase.
    zscore:
        The Geweke test threshold.

    Returns
    -------
    burn_in:
        Iteration where the first and the last fraction
        of the chain do not differ significantly
        regarding Geweke test.
    """
    nsamples, npar = chain.shape
    # number of fragments
    n = 20
    # round each element to the nearest integer
    # toward zero
    step = np.floor(nsamples / n).astype(int)
    fragments = np.arange(0, nsamples - 1, step)

    z = np.zeros((len(fragments), npar))
    for i, indices in enumerate(fragments):
        # Calculate z-score
        z[i, :], _ = calculate_zscore(chain[indices:, :])

    # Sort z-score for Bonferroni-Holm inverse
    # to sorting p-values
    max_z = np.max(np.absolute(z), axis=1)
    idxs = max_z.argsort()[::-1]  # sort descend
    alpha2 = zscore * np.ones((len(idxs)))

    for i in range(len(max_z)):
        alpha2[idxs[i]] = alpha2[idxs[i]] / (
            len(fragments) - np.where(idxs == i)[0] + 1
        )

    if np.any(alpha2 > max_z):
        burn_in = (np.where(alpha2 > max_z)[0][0]) * step
    else:
        burn_in = nsamples
        logger.warning(
            "Burn in index coincides with chain "
            "length. The chain seems to not have "
            "converged yet.\n"
            "You may want to use a larger number "
            "of samples."
        )

    return burn_in
